from typing import Union

from fgobjlib import FgObject


class FgFwPolicy(FgObject):
    """
    FgFwPolicy class represents FortiGate Firewall policy object and provides methods for validating parameters
    and generating both cli and api configuration data for use in external configuration applications.

    Attributes:
        policyid (int): Object ID
        srcintf (list):  Policy source interface(s), may be string or list of strings
        dstintf (list): Policy destination interface(s), may be string or list of strings
        srcaddr (list): Policy source address(es), may be string or list of strings
        dstaddr (list): Policy destination address(es), may be string or list of strings
        schedule (str): Policy schedule
        action (str):  Policy action, may be 'accept' or 'deny'
        logtraffic (str): Policy log action, may be 'utm', 'all' or 'disabled'
        nat (str):  Source NAT for policy.  ('enable', 'disable', or None=inherit)
        comment (str): Object comment
        vdom (str):  VDOM policy configured in  (if any)
        srcaddr_negate (str): Negate the source address in policy.  ('enable', 'disable', or None=inherit)
        dstaddr_negate (str): Negate the destination address in policy.  ('enable', 'disable', or None=inherit)
        service (str): Negate the service in policy.  ('enable', 'disable', or None=inherit)
    """

    def __init__(self, policyid: int = None, srcintf: Union[str, list] = None, dstintf: Union[str, list] = None,
                 srcaddr: Union[str, list] = None, dstaddr: Union[str, list] = None, service: Union[str, list] = None,
                 schedule: str = None, action: str = None, logtraffic: str = None, nat: str = None, vdom: str = None,
                 srcaddr_negate: str = None, dstaddr_negate: str = None, name: str = None, comment: str = None,
                 service_negate: str = None):
        """
        Args:
            policyid (int): ID of object.  Defines ID used in configs when API or CLI for config methods (default: 0)
            srcintf (list): string or list of strings referencing src interface(s) of policy
            dstintf (list): string or list of strings referencing dst interface(s) of policy
            srcaddr (list): string or list of strings referencing src address(s) of policy
            dstaddr (list): string or list of strings referencing dst address(s) of policy
            schedule (str): string referencing schedule to associated with policy (default: 'always')
            action (str): string sets action to assign to policy; may be 'accept' or 'deny' (default: deny)
            logtraffic (str): string set logtraffic action to assign; may be utm/all/disabled (default: disabled)
            nat (str): set nat enabled or disable ('enable', 'disable', or None=inherit)
            comment (str): set a comment up to 255 characters (default: None)
            vdom (str): opt - set vdom.  If unset object configs uses default fg context (default: None)
            srcaddr_negate (str): Negate the source address in policy.  ('enable', 'disable', or None=inherit)
            dstaddr_negate (str): Negate the destination address in policy.  ('enable', 'disable', or None=inherit)
            service (str): Negate the service in policy.  ('enable', 'disable', or None=inherit)
        """

        # Initialize the parent class - we do set this here, because the subclass will first verify obj_id
        # is acceptable for this class type in the above attribute set functions
        super().__init__(api='cmdb', api_path='firewall', api_name='policy', cli_path="config firewall policy",
                         obj_id=policyid, vdom=vdom)

        # Set parent class attributes
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'policyid': 'policyid', 'srcintf': 'srcintf', 'dstintf': 'dstintf', 'srcaddr': 'srcaddr',
                            'dstaddr': 'dstaddr', 'service': 'service', 'schedule': 'schedule', 'action': 'action',
                            'logtraffic': 'logtraffic', 'nat': 'nat', 'srcaddr_negate': 'srcaddr-negate',
                            'dstaddr_negate': 'dstaddr-negate', 'service_negate': 'service-negate', 'name': 'name'}

        self._cli_ignore_attrs = ['policyid']

        # Set Instance Variables (uses @property and setters defined below)
        self.policyid = policyid
        self.srcintf = srcintf
        self.dstintf = dstintf
        self.srcaddr = srcaddr
        self.dstaddr = dstaddr
        self.service = service
        self.schedule = schedule
        self.action = action
        self.logtraffic = logtraffic
        self.nat = nat
        self.srcaddr_negate = srcaddr_negate
        self.dstaddr_negate = dstaddr_negate
        self.service_negate = service_negate
        self.comment = comment
        self.name = name

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f', policyid={self.policyid}, srcintf={self.srcintf}, dstintf={self.dstintf}, ' \
                            f'srcaddr={self.srcaddr}, dstaddr={self.dstaddr}, service={self.service}, ' \
                            f'schedule={self.schedule}, action={self.action}, logtraffic={self.logtraffic}, ' \
                            f'nat={self.nat}, srcaddr_negate={self.srcaddr_negate}, ' \
                            f'dstaddr_negate={self.dstaddr_negate}, service_negate={self.service_negate}, ' \
                            f'comment={self.comment}, vdom={self.vdom}'

    # Static Methods
    @staticmethod
    def _validate_and_get_policy_obj(policy_object):
        """ Check the validity of policy objects and returns objects as list if valid

        Can be used to validate srcintf, dstintf, srcaddr, dstaddr and service objects

        Args:
            policy_object (list): string or list of strings containing srcintf(s)

        Returns:
            List
        """

        if policy_object is None:
            return policy_object

        else:
            # IF a single object was passed as a string, append it to intf_list else iterate the list and pull
            # out the strings of interfaces and append each to intf_list
            obj_list = []

            if isinstance(policy_object, str):
                obj_list.append({'name': policy_object})

            elif isinstance(policy_object, list):
                for item in policy_object:
                    obj_list.append({'name': item})

            else:
                raise Exception("'policy_object(s)', must be provided as string or list of strings")

            # Make sure each interface passed in is not all whitespace and it is less than 80 chars
            for item in obj_list:
                if isinstance(item['name'], str):
                    if item['name'].isspace():
                        raise Exception(f"{item} cannot be an empty string")

                    if not len(item['name']) < 80:
                        raise Exception("'policy_object(s)', must be 79 chars or less")

                else:
                    raise Exception("'policy_objects(s)' must be string or list of strings")

            # set self.<obj_type> attribute with the verified and formatted obj_list
            return obj_list

    # Instance Properties and Setters
    @property
    def policyid(self):
        return self._policyid

    @policyid.setter
    def policyid(self, policyid):
        """ Set self.policyid to policyid if policyid is valid or if not provided set policyid = 0.

        Args:
            policyid (int):  Integer representing ID of policy.

        Returns:
            None
        """
        if policyid is None:
            self._policyid = 0

        else:
            if isinstance(policyid, int):
                self._policyid = policyid
            else:
                raise Exception("'id', when set, must be an integer")

    @property
    def srcintf(self):
        return self._srcintf

    @srcintf.setter
    def srcintf(self, policy_object):
        """ Check the validity of srcintf objects and sets self.srcintf to list containing objects of dstintf

        Calls _validate_and_get_policy_obj()

        Args:
            policy_object (list): string or list of strings containing srcintfs(s)

        Returns:
            None
        """
        self._srcintf = self._validate_and_get_policy_obj(policy_object)

    @property
    def dstintf(self):
        return self._dstintf

    @dstintf.setter
    def dstintf(self, policy_object):
        """ Check the validity of dstintf objects and sets self.dstintf to list containing objects of dstintf

        Calls _validate_and_get_policy_obj()

        Args:
            policy_object (list): string or list of strings containing dstintf(s)

        Returns:
            None
        """
        self._dstintf = self._validate_and_get_policy_obj(policy_object)

    @property
    def srcaddr(self):
        return self._srcaddr

    @srcaddr.setter
    def srcaddr(self, policy_object):
        """ Check the validity of srcaddr objects and sets self.srcaddr to list containing objects of srcaddr

        Calls _validate_and_get_policy_obj()

        Args:
            policy_object (list): string or list of strings containing dstintf(s)

        Returns:
            None
        """
        self._srcaddr = self._validate_and_get_policy_obj(policy_object)

    @property
    def dstaddr(self):
        return self._dstaddr

    @dstaddr.setter
    def dstaddr(self, policy_object):
        """ Check the validity of dstaddr objects and sets self.dstaddr to list containing objects of dstaddr

        Calls _validate_and_get_policy_obj()

        Args:
            policy_object (list): string or list of strings containing dstintf(s)

        Returns:
            None
        """
        self._dstaddr = self._validate_and_get_policy_obj(policy_object)

    @property
    def service(self):
        return self._service

    @service.setter
    def service(self, policy_object):
        """ Check the validity of service objects and sets self.service to list containing objects of service

        Calls _validate_and_get_policy_obj()

        Args:
            policy_object (list): string or list of strings containing dstintf(s)

        Returns:
            None
        """
        self._service = self._validate_and_get_policy_obj(policy_object)

    @property
    def schedule(self):
        return self._schedule

    @schedule.setter
    def schedule(self, schedule):
        """ Set self.schedule to 'schedule' if 'schedule' name provided meets requirements

        Args:
            schedule (str): Policy schedule name

        Returns:
            None
        """
        if schedule is None:
            self._schedule = None

        else:
            if isinstance(schedule, str):
                if not len(schedule) <= 35:
                    raise ValueError("'schedule, when set, must be type str() 35 chars or less")

                if schedule.isspace():
                    raise ValueError("'schedule', when set, cannot be an empty string")
            else:
                raise ValueError("'schedule', when set, must be type str() 35 chars or less")

            self._schedule = schedule

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, action):
        """ Set self.action to 'action' if action meets requirements

        Args:
            action (str): policy action, may be 'accept' or 'deny'

        Returns:
            None
        """
        if action is None:
            self._action = None

        elif isinstance(action, str):
            if action.lower() == 'accept' or action.lower() == 'allow':
                self._action = 'accept'
            elif action.lower() == 'deny' or action.lower() == 'drop':
                self._action = 'deny'
            else:
                raise ValueError("'action', when set, must be type str() with value 'accept' or 'deny'")
        else:
            raise ValueError("'action', when set, must be type str() with value 'accept' or 'deny'")

    @property
    def logtraffic(self):
        return self._logtraffic

    @logtraffic.setter
    def logtraffic(self, logtraffic):
        """ Set self.logtraffic to 'logtraffic' if logtraffic is valid

        Args:
            logtraffic (str): Policy log action.  May be 'utm', 'all' or 'disabled'

        Returns:
            None
        """
        if logtraffic is None:
            self._logtraffic = None

        elif isinstance(logtraffic, str):
            if logtraffic.lower() == 'utm':
                self._logtraffic = 'utm'

            elif logtraffic.lower() == 'all':
                self._logtraffic = 'all'

            elif logtraffic.lower() == 'disabled' or logtraffic.lower() == 'disable':
                self._logtraffic = 'disabled'
            else:
                raise ValueError("'logtraffic', when set must be type str() with value 'utm', 'all' or 'disabled'")
        else:
            raise ValueError("'logtraffic', when set must be type str()")

    @property
    def nat(self):
        return self._nat

    @nat.setter
    def nat(self, nat):
        """ Set self.nat to 'nat' if valid.  True=enable, False=disable

        Args:
            nat (str): Policy source NAT. ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if nat is None:
            self._nat = None

        else:
            if isinstance(nat, str):
                if nat == 'enable':
                    self._nat = 'enable'
                elif nat == 'disable':
                    self._nat = 'disable'
                else:
                    raise ValueError("'nat', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'nat', when set, must be type str()")

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        """ Set self.comment to 'comment' if comment string within requirements

        Args:
            comment (str):  Comment for this policy object

        Returns:
            None
        """
        if comment is None:
            self._comment = None

        else:
            if isinstance(comment, str):
                if 1 <= len(comment) <= 1023:
                    self._comment = comment
                else:
                    raise ValueError("'comment', when set, must be type str() between 1 and 1023 chars")
            else:
                raise ValueError("'comment', when set, must be type str()")

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.comment to 'comment' if comment string within requirements

        Args:
            name (str):  Optional Name for this policy object

        Returns:
            None
        """
        if name is None:
            self._name = None

        else:
            if isinstance(name, str):
                if 1 <= len(name) <= 35:
                    self._name = name
                else:
                    raise ValueError("'name', when set, must be type str() between 1 and 1,023 chars")
            else:
                raise ValueError("'name', when set, must be type str()")

    @property
    def srcaddr_negate(self):
        return self._srcaddr_negate

    @srcaddr_negate.setter
    def srcaddr_negate(self, negate):
        """ Set the self.srcaddr_negate attribute representing negate type in policy

        Args:
            negate (str): ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if negate is None:
            self._srcaddr_negate = None

        else:
            if isinstance(negate, str):
                if negate == 'enable':
                    self._srcaddr_negate = 'enable'
                elif negate == 'disable':
                    self._srcaddr_negate = 'disable'
                else:
                    raise ValueError("'srcaddr_negate, when set, must be type str() with value 'enable' or 'disable")
            else:
                raise ValueError("'srcaddr_negate', when set, must be type str()")

    @property
    def dstaddr_negate(self):
        return self._dstaddr_negate

    @dstaddr_negate.setter
    def dstaddr_negate(self, negate):
        """ Set the self.dstaddr_negate attribute representing negate type in policy

        Args:
            negate (str): ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if negate is None:
            self._dstaddr_negate = None

        else:
            if isinstance(negate, str):
                if negate == 'enable':
                    self._dstaddr_negate = 'enable'
                elif negate == 'disable':
                    self._dstaddr_negate = 'disable'
                else:
                    raise ValueError("'dstaddr_negate', when set, must be type str() with value 'enable' or 'disable'")

            else:
                raise ValueError("'dstaddr_negate', when set, must be type str()")

    @property
    def service_negate(self):
        return self._service_negate

    @service_negate.setter
    def service_negate(self, negate):
        """ Set the self.service attribute representing negate type in policy

        Args:
            negate (str): ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if negate is None:
            self._service_negate = None

        else:
            if isinstance(negate, str):
                if negate == 'enable':
                    self._service_negate = 'enable'
                elif negate == 'disable':
                    self._service_negate = 'disable'
                else:
                    raise ValueError("'service_negate', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'service_negate', when set, must be type str()")



import ipaddress
from typing import Union

from fgobjlib import FgObject


class FgIpsecP1Interface(FgObject):
    """ FgIpsecP1Interface class represents FortiGate Firewall ipsec phase1 interface object and provides methods for
    validating parameters and generating both cli and api configuration data for use in external configuration
    applications

    Currently supports dynamic or static VPN using psk authentication. No support yet for advpn or mode-cfg

    Attributes:
        name (str): Name of ipsec phase1-interface object
        p1_type (str): Phase1-interface type
        interface (str): Name of locally attached fortigate interface
        proposal (list): Phase1-interface proposal(s)
        ike_version (int): ike version
        local_gw (str): Phase1-interface local-gw IP
        psksecret (str): Pre-shared key
        localid (str): Local ID
        remote_gw (str): Remote Gateway
        add_route (str):  add-route ('enable', 'disable', or None=inherit)
        add_gw_route (str): add-gw-route ('enable', 'disable', or None=inherit)
        keepalive (int): Keepalive in seconds
        net_device (str): net-device  ('enable', 'disable', or None=inherit)
        comment (str): phase1 comment
        vdom (str): Associated VDOM, if applicable
        tunnel_search (str):  tunnel-search ('next-hop', 'selectors' or None=inherit)
        dpd (str): phase1 DPD ('on-demand', 'on-idle', 'disable' or None=inherit)
        dhgrp (str): dhgrp
        nattraversal (str): nattraversal ('enable', 'disable', 'forced' or None=inherit)
        exchange_interface_ip (str): exchange-interface-ip ('enable', 'disable', or None=inherit)
    """

    def __init__(self, name: str = None, p1_type: str = None, interface: str = None, proposal: Union[str, list] = None,
                 ike_version: int = None, local_gw: str = None, psksecret: str = None, localid: str = None,
                 remote_gw: str = None, add_route: str = None, add_gw_route: str = None, keepalive: int = None,
                 net_device: str = None, comment: str = None, vdom: str = None, tunnel_search: str = None,
                 dpd: str = None, dhgrp: list = None, nattraversal: str = None, exchange_interface_ip: str = None):

        """
        Args:
            name (str): Set name of ipsec phase1-interface object
            p1_type (str): Set phase1-interface type
            interface (str): Set name of locally attached fortigate interface
            proposal (list): Set phase1-interface proposal(s)
            ike_version (int): Set ike version
            local_gw (str): Set phase1-interface local-gw IP
            psksecret (str): Set pre-shared key
            localid (str): Set local ID
            remote_gw (str): Set remote Gateway
            add_route (str):  Set add-route ('enable', 'disable', or None=inherit)
            add_gw_route (str): Set add-gw-route ('enable', 'disable', or None=inherit)
            keepalive (int): Set keepalive in seconds
            net_device (str): Set net-device ('enable', 'disable', or None=inherit)
            comment (str): Set phase1 comment
            vdom (str): Set associated VDOM, if applicable
            tunnel_search (str):  Set tunnel-search ('next-hop', 'selectors' or None=inherit)
            dpd (list): Set phase1 DPD ('on-demand', 'on-idle', 'disable' or None=inherit)
            nattraversal (str): Set nattraversal ('enable', 'disable', 'forced' or None=inherit)
            exchange_interface_ip (str): Set exchange-interfce-ip ('enable', 'disable', or None=inherit)
        """

        # Initialize the parent class
        super().__init__(api='cmdb', api_path='vpn.ipsec', api_name='phase1-interface',
                         cli_path="config vpn ipsec phase1-interface", obj_id=name, vdom=vdom)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name', 'p1_type': 'type', 'interface': 'interface', 'proposal': 'proposal',
                            'ike_version': 'ike-version', 'local_gw': 'local-gw', 'psksecret': 'psksecret',
                            'localid': 'localid', 'remote_gw': 'remote-gw', 'comment': 'comments',
                            'add_route': 'add-route', 'add_gw_route': 'add-gw-route', 'keepalive': 'keepalive',
                            'net_device': 'net-device', 'tunnel_search': 'tunnel-search', 'dpd': 'dpd',
                            'dhgrp': 'dhgrp', 'nattraversal': 'nattraversal',
                            'exchange_interface_ip': 'exchange-interface-ip'}

        self._cli_ignore_attrs = ['name']

        # Set instance attributes #
        self.name = name
        self.p1_type = p1_type
        self.interface = interface
        self.proposal = proposal
        self.ike_version = ike_version
        self.local_gw = local_gw
        self.psksecret = psksecret
        self.localid = localid
        self.remote_gw = remote_gw
        self.comment = comment
        self.add_route = add_route
        self.keepalive = keepalive
        self.add_gw_route = add_gw_route
        self.net_device = net_device
        self.tunnel_search = tunnel_search
        self.dpd = dpd
        self.dhgrp = dhgrp
        self.nattraversal = nattraversal
        self.exchange_interface_ip = exchange_interface_ip

        # Set object to string config
        self._obj_to_str += f', name={name}, p1_type={p1_type}, interface={self.interface}, ' \
                            f'proposal={self.proposal}, ike_version={self.ike_version}, local_gw={self.local_gw}, ' \
                            f'psksecret={self.psksecret}, localid={self.localid}, remote_gw={self.remote_gw}, ' \
                            f'comment={self.comment}, add_route={self.add_route}, add_gw_route={self.add_gw_route}, ' \
                            f'keepalive={self.keepalive}, net_device={self.net_device}, ' \
                            f'tunnel_search={self.tunnel_search}, dpd={self.dpd}, dhgrp={self.dhgrp},' \
                            f'nattraversal={self.nattraversal}, exchange_interface_ip={self.exchange_interface_ip}'

    # Instance properties and setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.name to name if name is valid

        Args:
            name: Name of object

        Returns:
            None
        """
        if name is None:
            self._name = None

        else:
            if name.isspace():
                raise Exception("'name', cannot be an empty string")
            if isinstance(name, str):
                if len(name) <= 35:
                    self._name = name
                else:
                    raise Exception("'name', must be type str() between 1 and 35 chars")
            else:
                raise Exception("'name', must be type str()")

    @property
    def proposal(self):
        return self._proposal

    @proposal.setter
    def proposal(self, proposal):
        """ Set self.proposal with list of proposals from proposal if items are all acceptable FG proposals

        Args:
            proposal (list): string containing a single p1 proposal or list of strings with one or more p1 propolsals

        Returns:
            None
        """
        valid_proposals = ['des-md5', 'des-sha', 'des-sha256', 'des-sha384', 'des-sha512', '3des-md5', '3des-sha1',
                           '3des-sha256', '3des-sha384', '3des-sha512', 'aes128-md5', 'aes128-sha1', 'aes128-sha256',
                           'aes128-sha384', 'aes128-sha512', 'aes192-md5', 'aes192-sha1', 'aes192-sha256',
                           'aes192-sha384', 'aes192-sha512', 'aes256-md5', 'aes256-sha1', 'aes256-sha256',
                           'aes256-sha384', 'aes256-sha512', 'aria128-md5', 'aria128-sha1', 'aria128-sha256',
                           'aria128-sha384', 'aria128-sha512', 'aria192-md5', 'aria192-sha1', 'aria192-sha256',
                           'aria192-sha384', 'aria192-sha512', 'aria256-md5', 'aria256-sha1', 'aria256-sha256',
                           'aria256-sha384', 'aria256-sha512', 'seed-md5', 'seed-sha1', 'seed-sha256', 'seed-sha384',
                           'seed-sha512']

        if proposal is None:
            self._proposal = None

        else:
            proposal_items = ''

            # IF a single object was passed as a string, append it to intf_list else iterate the list and pull
            # out the strings of interfaces and append each to intf_list
            if isinstance(proposal, str):

                # compare proposal to valid_proposals list
                if proposal in valid_proposals:
                    proposal_items += f"{proposal} "
                else:
                    raise ValueError("'proposal' provided is not a valid FortiGate phase1 proposal option")

            elif isinstance(proposal, list):
                for item in proposal:
                    if isinstance(item, str):

                        # compare proposal to valid proposals list
                        if item in valid_proposals:
                            proposal_items += f"{item} "
                        else:
                            raise ValueError("'proposal'(s) provided contain at least one that is not a valid "
                                             "FortiGate phase1 proposal option")
            else:
                raise ValueError("'proposal' must be type str() or list() of str()")

            self._proposal = proposal_items

    @property
    def dhgrp(self):
        return self._dhgrp

    @dhgrp.setter
    def dhgrp(self, dhgrp):
        """ Set self.dhgrp to dhgrp if dhgrp is valid ForitGate dhgrp

        The dhgrp may be passed in as a str, or a list.  If the dhgrp(s) passed in are valid, add each
        of those to a space separated values string and set in self.dhgrp

        Args:
            dhgrp (list): single int representing one dhgrp or a list of ints for one or more dhgrps

        Returns:
            None
        """
        if dhgrp is None:
            self._dhgrp = None

        else:
            dhgrp_items = ''
            valid_dhgrps = [1, 2, 5, 14, 15, 16, 17, 18, 19, 20, 21, 27, 28, 30, 31, 32]

            # IF a single object was passed as a string, append it to list else iterate the list and pull
            # out the dhgrps and add to list to be set as self object
            if isinstance(dhgrp, int):
                # compare proposal to valid_list
                if dhgrp in valid_dhgrps:
                    dhgrp_items += "{} ".format(dhgrp)
                else:
                    raise ValueError("'dhgrp' provided is not a valid fortigate dhgrp option")

            elif isinstance(dhgrp, list):
                for item in dhgrp:
                    if isinstance(item, int):

                        # compare proposal to valid proposals list
                        if item in valid_dhgrps:
                            dhgrp_items += "{} ".format(item)
                        else:
                            raise ValueError(f"At least one 'dhgrp' provided is not a valid fortigate dhgrp option")
            else:
                raise ValueError("dhgrp must be type int()")

            self._dhgrp = dhgrp_items

    @property
    def p1_type(self):
        return self._p1_type

    @p1_type.setter
    def p1_type(self, p1_type):
        """ Set self.pt_type to p1_type if p1_type is valid

        Args:
            p1_type (str): Phase1-interface type.  ('dynamic', 'static', 'ddns' or None=inherit)

        Returns:

        """
        if p1_type is None:
            self._p1_type = None
        else:
            if isinstance(p1_type, str):
                if p1_type.lower() == 'dynamic':
                    self._p1_type = 'dynamic'
                elif p1_type.lower() == 'static':
                    self._p1_type = 'static'
                elif p1_type.lower() == 'ddns':
                    raise ValueError("p1_type of 'ddns' is not yet supported")
                else:
                    raise ValueError("'p1_type', must be type str() with value 'dynamic' or 'static'")
            else:
                raise ValueError("'p1_type', must be type str() with value of 'dynamic' or 'static'")

    @property
    def interface(self):
        return self._interface

    @interface.setter
    def interface(self, interface):
        """ set self.local_intfs to intf if intfs is valid

        Args:
            interface (str): Local interface for p1 attachment  (1 to 35 chars)

        Returns:
            None
        """
        if interface is None:
            self._interface = None

        else:
            if interface.isspace():
                raise ValueError("'interface', cannot be an empty string")
            if isinstance(interface, str):
                if len(interface) < 35:
                    self._interface = interface
                else:
                    raise ValueError("'interface', when set, must be type str() between 1 and 35 chars")
            else:
                raise ValueError("'interface', when set, must be type str()")

    @property
    def ike_version(self):
        return self._ike_version

    @ike_version.setter
    def ike_version(self, ike_version):
        """ Set self.ike_version to 1 or 2 if ike_version = 1 or 2.  Otherwise raise Exception

        Args:
            ike_version (int): ike-version.  (1 or 2)

        Returns:
            None
        """
        if ike_version is None:
            self._ike_version = None
        else:
            if isinstance(ike_version, int):
                if ike_version == 1:
                    self._ike_version = 1
                elif ike_version == 2:
                    self._ike_version = 2
                else:
                    raise ValueError("'ike_version', must be type int() with value '1' or '2'")
            else:
                raise ValueError("'ike_version' must be type int() with value = '1' or '2'")

    @property
    def local_gw(self):
        return self._local_gw

    @local_gw.setter
    def local_gw(self, local_gw):
        """ Set self.local_gw to local_gw if local_gw is valid ipv4 address

        Args:
            local_gw (str): Local gateway.  (valid ipv4 address as str())

        Returns:
            None
        """
        if local_gw is None:
            self._local_gw = None
        else:
            try:
                self._local_gw = str(ipaddress.ip_address(local_gw))
            except ValueError:
                raise ValueError("'local_gw', when set, must type str() with value containing a valid ipv4 address")

    @property
    def remote_gw(self):
        return self._remote_gw

    @remote_gw.setter
    def remote_gw(self, remote_gw):
        """ Set self.remote_gw to remote_gw if remote_gw is valid ipv4 address

        Args:
            remote_gw (str): Address of remote vpn peer gateway.  (valid ipv4 address as str())

        Returns:
            None
        """
        if remote_gw is None:
            self._remote_gw = None
        else:
            try:
                self._remote_gw = str(ipaddress.ip_address(remote_gw))
            except ValueError:
                raise ValueError("'remote_gw', when set, must be type str() with value containing a valid ipv4 address")

    @property
    def psksecret(self):
        return self._psksecret

    @psksecret.setter
    def psksecret(self, psk):
        """ Set self.psk to psk if psk valid

        Args:
            psk (str): Phase1 psksecret.  (6 to 30 chars)

        Returns:
            None
        """
        if psk is None:
            self._psksecret = None
        else:
            if isinstance(psk, str):
                if 6 <= len(psk) <= 30:
                    self._psksecret = psk
                else:
                    raise ValueError("'psksecret', must be type str() between 6 and 30 chars")
            else:
                raise ValueError("'psksecret', must be type str()")

    @property
    def localid(self):
        return self._localid

    @localid.setter
    def localid(self, localid):
        """ Set self.local_id to local_id if local_id is valid

        Args:
            localid (str): Phase1 local id.  (up to 68 chars)

       Returns:
            None
        """
        if localid is None:
            self._localid = None
        else:
            if isinstance(localid, str):
                if 1 <= len(localid) <= 63:
                    self._localid = localid
                else:
                    raise ValueError("'localid', when set, must be type str() between 1 and 63 chars")
            else:
                raise ValueError("'localid', when set, must be type str()")

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        """ Set self.comment to comment if comment is valid

        Args:
            comment (str): Phase1 comment.  (up to 1023 chars)

        Returns:
            None
        """
        if comment is None:
            self._comment = None
        else:
            if isinstance(comment, str):
                if 1 <= len(comment) <= 1023:
                    self._comment = comment
                else:
                    raise ValueError("'description', when set, must be type str() between 1 and 1,023 chars")
            else:
                raise Exception("'description', when set, must be type str()")

    @property
    def keepalive(self):
        return self._keepalive

    @keepalive.setter
    def keepalive(self, keepalive):
        """ Set self.keepalive if keepalive valid

        Args:
            keepalive (int): phase1 keepalive  (10-900)

        Returns:
            None
        """
        if keepalive is None:
            self._keepalive = None
        else:
            if isinstance(keepalive, int):
                if 10 <= keepalive <= 900:
                    self._keepalive = keepalive
                else:
                    raise ValueError("'keepalive', when set, must be type int() between 10 and 900")
            else:
                raise ValueError("'keepalive', when set, must be type int()")

    @property
    def add_route(self):
        return self._add_route

    @add_route.setter
    def add_route(self, add_route):
        """ Set self.add_route

        Args:
            add_route (str): add-route.  ('enable', 'disable' or None=inherit)

        Returns:

        """
        if add_route is None:
            self._add_route = None
        else:
            if isinstance(add_route, str):
                if add_route == 'enable':
                    self._add_route = 'enable'
                elif add_route == 'disable':
                    self._add_route = 'disable'
                else:
                    raise ValueError("'add_route', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'add_route', when set, must be type str()")

    @property
    def add_gw_route(self):
        return self._add_gw_route

    @add_gw_route.setter
    def add_gw_route(self, add_gw_route):
        """ Set self.add_gw_route

        Args:
            add_gw_route (str): add-gw-route. ('enable', 'disable' or None=inherit)

        Returns:
            None
        """
        if add_gw_route is None:
            self._add_gw_route = None
        else:
            if isinstance(add_gw_route, str):
                if add_gw_route == 'enable':
                    self._add_gw_route = 'enable'
                elif add_gw_route == 'disable':
                    self._add_gw_route = 'disable'
                else:
                    raise ValueError("'add_gw_route', when set, must be type str() with value 'enable' or 'disable")
            else:
                raise ValueError("'add_gw_route', when set, must be type str()")

    @property
    def net_device(self):
        return self._net_device

    @net_device.setter
    def net_device(self, net_device):
        """ set self.net_device

        Args:
            net_device (str): net-device. ('enable', 'disable' or None=inherit)

        Returns:
            None
        """
        if net_device is None:
            self._net_device = None
        else:
            if isinstance(net_device, str):
                if net_device == 'enable':
                    self._net_device = 'enable'
                elif net_device == 'disable':
                    self._net_device = 'disable'
                else:
                    raise ValueError("'net_device', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'net_device', when set, must be type str()")

    @property
    def tunnel_search(self):
        return self._tunnel_search

    @tunnel_search.setter
    def tunnel_search(self, tunnel_search):
        """ Set self.tunnel_search

        Args:
            tunnel_search (str): tunnel-search.  ('selectors', 'nexthop', None=inherit)

        Returns:
            None
        """
        if tunnel_search is None:
            self._tunnel_search = None
        else:
            if isinstance(tunnel_search, str):
                if tunnel_search.lower() == 'selectors':
                    self._tunnel_search = 'selectors'
                elif tunnel_search.lower() == 'nexthop':
                    self._tunnel_search = 'nexthop'
                else:
                    raise Exception("'tunnel_search' when set, must be type str() with value 'selectors', 'nexthop'")
            else:
                raise ValueError("'tunnel_search', when set, must be type str(0")

    @property
    def dpd(self):
        return self._dpd

    @dpd.setter
    def dpd(self, dpd):
        """ Set self.dpd

        Args:
            dpd (str): phase1 dpd.   ('disable', 'on-idle', 'on-demand', None=inherit)

        Returns:
            None
        """
        if dpd is None:
            self._dpd = None
        else:
            if isinstance(dpd, str):
                if dpd.lower() == 'disable':
                    self._dpd = 'disable'
                elif dpd.lower() == 'on-idle':
                    self._dpd = 'on-idle'
                elif dpd.lower == 'on-demand':
                    self._dpd = 'on-demand'
                else:
                    raise Exception("'dpd', when set, must be type str() with value 'disable', 'on-idle' or "
                                    "'on-demmand")
            else:
                raise ValueError("'dpd', when set, must be type str()")

    @property
    def nattraversal(self):
        return self._nattraversal

    @nattraversal.setter
    def nattraversal(self, nattraversal):
        """ Set self.nat_traversal

        Args:
            nattraversal (str): nat-traversal.  ('enable', 'disable', 'forced', None=inherit)

        Returns:
            None
        """
        if nattraversal is None:
            self._nattraversal = None
        else:
            if isinstance(nattraversal, str):
                if nattraversal.lower() == 'enable':
                    self._nattraversal = 'enable'
                elif nattraversal.lower() == 'disable':
                    self._nattraversal = 'disable'
                elif nattraversal.lower() == 'forced':
                    self._nattraversal = 'forced'
                else:
                    raise ValueError("'nattraversal', when set, must be type str() with value 'enable', 'disable' "
                                     "or 'forced'")
            else:
                raise ValueError("'nattraversal', when set, must be type str()")

    @property
    def exchange_interface_ip(self):
        return self._exchange_interface_ip

    @exchange_interface_ip.setter
    def exchange_interface_ip(self, exchange_interface_ip):
        """ Set self.exchange_interface_ip

        Args:
            exchange_interface_ip (str): exchange-interface-ip. ('enable', 'disable' or None)

        Returns:
            None
        """
        if exchange_interface_ip is None:
            self._exchange_interface_ip = None
        else:
            if isinstance(exchange_interface_ip, str):
                if exchange_interface_ip == 'enable':
                    self._exchange_interface_ip = 'enable'
                elif exchange_interface_ip == 'disable':
                    self._exchange_interface_ip = 'disable'
                else:
                    raise ValueError("exchange_interface_ip, when set, must be type str() with value 'enable' or "
                                     "'disable")
            else:
                raise ValueError("exchange_interface_ip, when set, must be type str()")



from fgobjlib import FgObject

class FgVdom(FgObject):
    """ FGVdom class represents a FortiGate VDOM object and provides methods for validating parameters
    and generating both cli and api configuration data for use in external configuration applications

    Note: VDOMs must be enabled on FortiGate prior to using results from this object for configuring FortiGates

    Attributes:
        name (str): Name of vdom-link object
    """

    def __init__(self, name: str = None):
        """
        Args:
            name (str): Name of vdom
        """

        # Initialize the parent class
        super().__init__(api='cmdb', api_path='system', api_name='vdom', cli_path="config vdom",
                         vdom='global', obj_id=name)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name'}
        self._cli_ignore_attrs = ['name']

        # Enable global config instead of per VDOM
        self.is_global = True

        # Set instance attributes
        self.name = name

        # Set obj to str conversion
        self._obj_to_str += f', name={self.name}'

    # Instance properties and setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.name to name if name is valid

        Args:
            name (str): Name of vdom

        Returns:
            None
        """
        if name:
            if name.isspace():
                raise ValueError("'name', cannot be an empty string")
            for char in name:
                if char.isspace():
                    raise ValueError("'name' may not contain spaces")
            if isinstance(name, str):
                if 1 <= len(name) <= 31:
                    self._name = name
                else:
                    raise ValueError("'name', must be type str() between 1 and 31 chars")
            else:
                raise ValueError("'name', must be type str() between 1 and 31 chars")
        else:
            self._name = None



from typing import Union

from fgobjlib import FgObject


class FgFwAddressGroup(FgObject):
    """
    FgFwAddressGroup class represents FortiGate Firewall firewall address group object and provides methods for
    validating parameters and generating both cli and api configuration data for use in external configuration
    applications.

    Attributes:
        name (str): Name of addrgrp to be created
        member (list): List of member fw addresses to include in group
        exclude (str): Set exclude addresses from addrgrp 'enable' or 'disable'
        exclude_member (list): The members to exclude from addrgrp
        comment (str):  Comment for addrgrp
        visibility (str): Set gui visibility 'enable' or 'disable'
        allow_routing (str): Set allow addrgrp use in static routing configuration 'enable' or 'disable'
    """

    def __init__(self, name: str = None, member: Union[str, list] = None, exclude: str = None,
                 exclude_member: Union[str, list] = None, comment: str = None, visibility: str = None,
                 allow_routing: str = None, vdom: str = None):
        """
        Args:
            name (str): Name of addrgrp to be created
            member (list): List of member fw addresses to include in group
            exclude (str): Set exclude addresses from addrgrp 'enable' or 'disable'
            exclude_member (list): The members to exclude from addrgrp
            comment (str):  Comment for addrgrp
            visibility (str): Set gui visibility 'enable' or 'disable'
            allow_routing (str): Set allow addrgrp use in static routing configuration 'enable' or 'disable'
        """

        # Initialize the parent class - we do set this here, because the subclass will first verify obj_id
        # is acceptable for this class type in the above attribute set functions
        super().__init__(api='cmdb', api_path='firewall', api_name='addrgrp', cli_path="config firewall addrgrp",
                         obj_id=name, vdom=vdom)

        # Set parent class attributes
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name', 'member': 'member', 'exclude': 'exclude',
                            'exclude_member': 'exclude-member', 'comment': 'comment', 'visibility': 'visibility',
                            'allow_routing': 'allow-routing'}

        self._cli_ignore_attrs = []

        # Set Instance Variables (uses @property and setters defined below)
        self.name = name
        self.member = member
        self.exclude = exclude
        self.exclude_member = exclude_member
        self.visibility = visibility
        self.allow_routing = allow_routing
        self.comment = comment

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f', name={self.name}, member={self.member}, exclude={self.exclude}, ' \
                            f'exclude_member={self.exclude_member}, visibility={self.visibility}, ' \
                            f'allow_routing={self.allow_routing}, comment={self.comment}'

    # Static Methods
    @staticmethod
    def _validate_and_get_members(members):
        """ Check the validity of policy objects and returns objects as list if valid

        Can be used to validate srcintf, dstintf, srcaddr, dstaddr and service objects

        Args:
            members (list): string or list of strings containing fw address objs as members

        Returns:
            List
        """

        if members is None:
            return None
        else:
            # IF a single object was passed as a string, append it to member_list else iterate the list and pull
            # out the strings of members and append each to member_list
            member_list = []

            if isinstance(members, str):
                member_list.append({'name': members})

            elif isinstance(members, list):
                for item in members:
                    member_list.append({'name': item})

            else:
                raise Exception("'members', must be provided as string or list of strings")

            # Make sure each interface passed in is not all whitespace and it is less than 80 chars
            for item in member_list:
                if isinstance(item['name'], str):
                    if item['name'].isspace():
                        raise Exception("members cannot be an contain values that are an empty string")

                    if not len(item['name']) < 80:
                        raise Exception("'policy_object(s)', must be 79 chars or less")

                else:
                    raise Exception("'policy_objects(s)' must be string or list of strings")

            # set self.<obj_type> attribute with the verified and formatted obj_list
            return member_list

    # Instance Properties and Setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.comment to 'comment' if comment string within requirements

        Args:
            name (str):  Optional Name for this policy object

        Returns:
            None
        """
        if name is None:
            self._name = None

        else:
            if isinstance(name, str):
                if 1 <= len(name) <= 79:
                    self._name = name
                else:
                    raise ValueError("'name', when set, must be type str() between 1 and 79 chars")
            else:
                raise ValueError("'name', when set, must be type str()")

    @property
    def member(self):
        return self._member

    @member.setter
    def member(self, member):
        """ Check the validity of members list

        Calls _validate_and_get_members()

        Args:
            member(list): string or list of strings containing member fw address object names

        Returns:
            None
        """
        self._member = self._validate_and_get_members(member)

    @property
    def exclude_member(self):
        return self._exclude_member

    @exclude_member.setter
    def exclude_member(self, exclude_member):
        """ Check the validity of members list

        Calls _validate_and_get_members()

        Args:
            exclude_member (list): string or list of strings containing member fw address object names

        Returns:
            None
        """
        self._exclude_member = self._validate_and_get_members(exclude_member)

    @property
    def visibility(self):
        return self._visibility

    @visibility.setter
    def visibility(self, visibility):
        """ Set self.visibility if valid.  allowed values: 'enable' or 'disable'

        Args:
            visibility (str): addrgrp visibility. ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if visibility is None:
            self._visibility = None

        else:
            if isinstance(visibility, str):
                if visibility == 'enable':
                    self._visibility = 'enable'
                elif visibility == 'disable':
                    self._visibility = 'disable'
                else:
                    raise ValueError("'visibility', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'visibility', when set, must be type str()")

    @property
    def allow_routing(self):
        return self._allow_routing

    @allow_routing.setter
    def allow_routing(self, allow_routing):
        """ Set self.allow_routing  if valid.  allowed values: 'enable' or 'disable'

        Args:
            allow_routing (str): allow object in routing config. ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if allow_routing is None:
            self._allow_routing = None

        else:
            if isinstance(allow_routing, str):
                if allow_routing == 'enable':
                    self._allow_routing = 'enable'
                elif allow_routing == 'disable':
                    self._allow_routing = 'disable'
                else:
                    raise ValueError("'allow_routing', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'allow_routing', when set, must be type str()")

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        """ Set self.comment to 'comment' if comment string within requirements

        Args:
            comment (str):  Comment for this policy object

        Returns:
            None
        """
        if comment is None:
            self._comment = None

        else:
            if isinstance(comment, str):
                if 1 <= len(comment) <= 255:
                    self._comment = comment
                else:
                    raise ValueError("'comment', when set, must be type str() between 1 and 255 chars")
            else:
                raise ValueError("'comment', when set, must be type str()")



from .fg_object import FgObject
from .fg_sys_router_static import FgRouteIPv4
from .fg_sys_interface import FgInterfaceIpv4
from .fg_fw_policy import FgFwPolicy
from .fg_fw_address import FgFwAddress
from .fg_fw_addrgrp import FgFwAddressGroup
from .fg_fw_service import FgFwService
from .fg_vpn_ipsec_p1_interface import FgIpsecP1Interface
from .fg_vpn_ipsec_p2_interface import FgIpsecP2Interface
from .fg_sys_vdomlink import FgVdomLink
from .fg_sys_vdom import FgVdom


from fgobjlib import FgObject


class FgVdomLink(FgObject):
    """ FgVdomLink class represents FortiGate system vdom-link object and provides methods for validating parameters
    and generating both cli and api configuration data for use in external configuration applications

    Attributes:
        name (str): Name of vdom-link object
        vdom_enabled (bool):  Vdom enabled on target object True or False
    """

    def __init__(self, name: str = None, vlink_type: str = None, vdom_enabled: bool = None):
        """
        Args:
            name (str): Name of vdom-link object
            vdom_enabled (bool): VDOMs enabled on target FortiGate?  Set to True if VDOMS enabled False if not.
        """

        # Initialize the parent class
        super().__init__(api='cmdb', api_path='system', api_name='vdom-link', cli_path="config system vdom-link",
                         vdom='global', obj_id=name)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name', 'vlink_type': 'type'}
        self._cli_ignore_attrs = []

        if vdom_enabled is True:
            self.vdom_enabled = True

        # Set instance attributes
        self.name = name
        self.vlink_type = vlink_type

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f', name={self.name}, vlink_type={self.vlink_type}'

    # Instance properties and setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.name to name if name is valid

        Args:
            name (str): Name of vdom-link

        Returns:
            None
        """
        if name is None:
            self._name = None

        else:
            if name.isspace():
                raise ValueError("'name', cannot be an empty string")
            if isinstance(name, str):
                if 1 <= len(name) <= 11:
                    self._name = name
                else:
                    raise ValueError("'name', must be type str() between 1 and 11 chars")
            else:
                raise ValueError("'name', must be type str()")

    @property
    def vlink_type(self):
        return self.vlink_type

    @vlink_type.setter
    def vlink_type(self, vlink_type):
        if vlink_type is None:
            self._vlink_type = None

        else:
            if isinstance(vlink_type, str):
                if vlink_type.lower() == 'ppp':
                    self._vlink_type = 'ppp'
                elif vlink_type.lower() == 'ethernet' or vlink_type == 'eth':
                    self._vlink_type = 'ethernet'
                else:
                    raise ValueError("'vlink_type', when set, must be type str with values either 'ppp' or 'ethernet'")
            else:
                raise ValueError("'vlink_type', when set, must be type str()")



import ipaddress

from fgobjlib import FgObject


class FgFwAddress(FgObject):
    """
    FgFwAddress  represents FortiGate firewall address object and provides methods for validating parameters
    and generating both cli and api configuration data for use in external configuration applications.

    Currently supports address types subnet (default), iprange and fqdn

    Attributes:
        name (str): Name of instance object
        type (str): Type of address object
        subnet (str): Subnet for "ipmask" type object
        fqdn (str): Fqdn for "fqdn" type object
        start_ip (str): start-ip for "iprange" type object
        end_ip (str): end-ip for "iprange: type object
        visibility (str): Object visibility ('enable', 'disable', or None=inherit)
        comment (str): Object comment
    """

    def __init__(self, name: str = None, type: str = None, subnet: str = None, fqdn: str = None,
                 start_ip: str = None, end_ip: str = None, visibility: str = None, associated_interface: str = None,
                 vdom: str = None, comment: str = None):
        """
        Args:
            name (str): required - Name of object.  Defines name used in configs when API or CLI for config methods
            type (str): optional - Set type of address object.  (default: ipmask)
            subnet (str): optional - Set subnet for address object of type ipmask. (default: None)
            fqdn (str): optional - Set fqdn for address object of type fqdn. (default: None)
            start_ip (str): optional - Set start-ip for address object of type iprange. (default: None)
            end_ip (str): optional - Set end-ip address object of type iprange. (default: None)
            visibility (bool): optional - Set visibility option to True-[enabled] or False-[disabled]  (default: None)
            associated_interface (str): optional - Set associated-interface for display in FortiGate GUI (default: None)
            comment (str): optional - Set a comment up to 255 characters (default: None)
            vdom (str): optional - Set vdom.  If unset object configs uses default fg context (default: None)
        """

        # Initialize the parent class
        super().__init__(api='cmdb', api_path='firewall', api_name='address',  cli_path="config firewall address",
                         obj_id=name, vdom=vdom)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name', 'type': 'type', 'subnet': 'subnet', 'fqdn': 'fqdn',
                            'associated_interface': 'associated-interface', 'visibility': 'visibility',
                            'comment': 'comments', 'start_ip': 'start-ip', 'end_ip': 'end-ip'}

        self._cli_ignore_attrs = []

        # Set instance attributes
        self.name = name
        self.type = type
        self.subnet = subnet
        self.fqdn = fqdn
        self.visibility = visibility
        self.associated_interface = associated_interface
        self.comment = comment
        self.start_ip = start_ip
        self.end_ip = end_ip

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f', name={self.name}, type={self.type}, subnet={self.subnet}, fqdn={self.fqdn}, ' \
                            f'start_ip={self.start_ip}, visibility={self.visibility}, ' \
                            f'associated_interface={self.associated_interface}, comment={self.comment}'

    # Instance Properties and Setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """  Set self.name attribute to name if name provided is valid for FG object

        Args:
            name (str): Name of firewall address object.

        Returns:
            None
        """

        if name is None:
            self._name = None

        else:
            if isinstance(name, str):
                if name.isspace():
                    raise ValueError("'name', cannot be an empty string")
                if 1 <= len(name) <= 79:
                    self._name = name
                else:
                    raise ValueError("'name', must be type str() 79 chars or less")
            else:
                raise ValueError("'name', must be type str()")

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        """ Set self.type attribute to type if type provided is valid for FG object

        Args:
            type (str): Type of firewall object, may be 'ipmask', 'iprange', 'fqdn' or None (None = FG Default)

        Returns:
            None
        """
        if type is None:
            self._type = None

        else:
            if isinstance(type, str):
                if type in ['ipmask', 'iprange', 'fqdn']:
                    self._type = type
                else:
                    raise ValueError("'type', when set, must be type str() with values 'ipmask', 'iprange "
                                     "or 'fqdn'")
            else:
                raise ValueError("'type', when set, must be type str()")

    @property
    def subnet(self):
        return self._subnet

    @subnet.setter
    def subnet(self, subnet):
        """ Set self.subnet to subnet if subnet is valid ipv4 network/mask

        Args:
            subnet (str): valid ipv4 network/mask, for use when self's 'type' is ipmask or None (which is default type)

        Returns:
            None
        """
        if subnet is None:
            self._subnet = None
        else:
            if isinstance(subnet, str):
                try:
                    self._subnet = str(ipaddress.ip_network(subnet))
                except ValueError:
                    raise ValueError("'subnet', when specified must be type str with value of a valid ipv4 address")
            else:
                raise ValueError("'subnet', when set, must be type str() with value of a valid ipv4 address")

    @property
    def fqdn(self):
        return self._fqdn

    @fqdn.setter
    def fqdn(self, fqdn):
        """ Set self.fqdn to fqdn if fqdn provided meets FortiGate requirements for this parameter

        Args:
            fqdn (str): fqdn for use when self type is also set to 'fqdn'

        Returns:
            None
        """
        if fqdn is None:
            self._fqdn = None

        elif isinstance(fqdn, str):
            if 1 <= len(fqdn) <= 255:
                self._fqdn = fqdn
            else:
                raise ValueError("'fqdn', when set, must be type str() between 1 and 255 chars")
        else:
            raise ValueError("'fqdn', when set, must be type str()")

    @property
    def visibility(self):
        return self._visibility

    @visibility.setter
    def visibility(self, visibility):
        """ Set self.visibility

        Args:
            visibility (str): object visibility in GUI.  ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if visibility is None:
            self._visibility = None
        else:
            if isinstance(visibility, str):
                if visibility == 'enable':
                    self._visibility = 'enable'
                elif visibility == 'disable':
                    self._visibility = 'disable'
                else:
                    raise ValueError("'visibility', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'visibility', when set, must be type str()")

    @property
    def associated_interface(self):
        return self._associated_interface

    @associated_interface.setter
    def associated_interface(self, intf):
        """ Set self.associated_interface if associated_interface provided meets requirements

        Args:
            intf (str): Name of associated interface for GUI display.  (1-35 chars)

        Returns:
            None
        """
        if intf is None:
            self._associated_interface = None

        else:
            if intf.isspace():
                raise ValueError("'associated_interface', when set, cannot be an empty string")
            if isinstance(intf, str):
                if 1 <= len(intf) <= 35:
                    self._associated_interface = intf
                else:
                    raise ValueError("'associated_interface', when set, must be type str() between 1 and 35 chars")
            else:
                raise ValueError("'associated_interface', when set, must be type str()")

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        """ Set self.comment if comment provided meets requirements

        Args:
            comment (str): Comment for this FG object

        Returns:
            None
        """
        if comment is None:
            self._comment = None

        else:
            if isinstance(comment, str):
                if 1 <= len(comment) <= 255:
                    self._comment = comment
                else:
                    raise ValueError("'comment', when set, must be type str() between 1 and 1,023 chars")
            else:
                raise ValueError("'comment', when set, must be type str()")

    @property
    def start_ip(self):
        return self._start_ip

    @start_ip.setter
    def start_ip(self, start_ip):
        """ Set self.start_ip

       Validates that start_ip, if set is a valid ipv4 address

        Args:
            start_ip (str):  Valid IPv4 address for FortiOS start-ip

        Returns:
            None
        """
        if start_ip is None:
            self._start_ip = None
        else:
            if isinstance(start_ip, str):
                try:
                    self._start_ip = str(ipaddress.ip_address(start_ip))
                except ValueError:
                    raise ValueError("'start_ip', when set must be type str() with value as valid ipv4 address")
            else:
                raise ValueError("'start_ip', when set must be type str() with value a valid ipv4 address")

    @property
    def end_ip(self):
        return self._end_ip

    @end_ip.setter
    def end_ip(self, end_ip):
        """ Set self.end_ip if end_ip is a valid ipv4 address or None

        Args:
            end_ip (str):  Valid IPv4 address for FortiOS end-ip

        Returns:
            None
        """
        if end_ip is None:
            self._end_ip = None
        else:
            if isinstance(end_ip, str):
                try:
                    self._end_ip = str(ipaddress.ip_address(end_ip))
                except ValueError:
                    raise ValueError("'end_ip', when set must be type str() with value as valid ipv4 address")
            else:
                raise ValueError("'end_ip', when set must be type str() with value a valid ipv4 address")



import re
from typing import Union

from fgobjlib import FgObject


class FgFwService(FgObject):
    """
    FgFwService class represents FortiGate Firewall service custom object and provides methods for validating parameters
    and generating both cli and api configuration data for use in external configuration applications

    Currently only supports tcp/udp/sctp and icmp type objects

    Currently, does not support proxy objects or setting of iprange, fqdn, tcp timers (halfopen/halfclose/timewait),
    or check-reset-range

    Attributes:
            name (str): Name of this FG object
            vdom: (str): Vdom associated with this FG object
            protocol (str): Protocol type for this FG service object
            tcp_portrange (list): TCP port range. May be string or list of strings.  ex. ['80', '1024-65553']
            udp_portrange (list): UDP port range.  May be string or list of strings.  ex. ['80', '1024-65553']
            sctp_portrange (list): SCTP port range.  May be string or list of strings. ex. ['80', '1024-65553']
            protocol_number (int):  IP protocol Number
            comment (str): Comment for this object
            visibility (str): Visibility of this object in GUI.  ('enable', 'disable', or None=inherit)
            session_ttl (int):  Session TTL for TCP sessions
            udp_idle_timer (int): Idle timer setting for UDP sessions
            category (str): Category to assign service object to in FG
            icmptype (int): Value of icmp type.  Used when self's protocol is 'icmp'
            icmpcode (int): Value of icmp type.  Used when self's protocol is 'icmp'
    """

    def __init__(self, name: str = None, vdom: str = None, protocol: str = None, tcp_portrange: Union[str, list] = None,
                 udp_portrange: Union[str, list] = None, sctp_portrange: Union[str, list] = None,
                 protocol_number: int = None, comment: str = None, visibility: str = None, session_ttl: int = None,
                 udp_idle_timer: int = None, category: str = None, icmptype: int = None, icmpcode: int = None):
        """
        Args:
            name (str): Name of this FG object
            vdom: (str): Vdom associated with this FG object
            protocol (str): Protocol type for this FG service object
            tcp_portrange (list): TCP port range. May be string or list of strings.  ex. ['80', '1024-65553']
            udp_portrange (list): UDP port range.  May be string or list of strings.  ex. ['80', '1024-65553']
            sctp_portrange (list): SCTP port range.  May be string or list of strings. ex. ['80', '1024-65553']
            protocol_number (int): IP Protocol Number. Use when self's protocol set to 'ip'.
            comment (str): Comment for this object
            visibility (str): Visibility of this object in GUI.  ('enable', 'disable', or None=inherit)
            session_ttl (int):  Session TTL for TCP sessions
            udp_idle_timer (int): Idle timer setting for UDP sessions
            category (str): Category to assign service object to in FG
            icmptype (int): Value of icmp type.  Used when self's protocol is 'icmp'
            icmpcode (int): Value of icmp code.  Used when self's protocol is 'icmp'
        """

        # Initialize the parent class
        super().__init__(api='cmdb', api_path='firewall.service', api_name='custom',
                         cli_path="config firewall service custom", obj_id=name, vdom=vdom)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name',  'protocol': 'protocol', 'tcp_portrange': 'tcp-portrange',
                            'udp_portrange': 'udp-portrange', 'sctp_portrange': 'sctp-portrange',
                            'icmptype': 'icmptype', 'comment': 'comments', 'visibility': 'visibility',
                            'session_ttl': 'session-ttl', 'udp_idle_timer': 'udp-idle-timer', 'category': 'category',
                            'protocol_number': 'protocol-number', 'icmpcode': 'icmpcode'}

        # Set attributes to ignore on CLI based configuration
        self._cli_ignore_attrs = []

        # Set instance attributes
        self.name = name
        self.protocol = protocol
        self.tcp_portrange = tcp_portrange
        self.udp_portrange = udp_portrange
        self.sctp_portrange = sctp_portrange
        self.protocol_number = protocol_number
        self.comment = comment
        self.visibility = visibility
        self.session_ttl = session_ttl
        self.udp_idle_timer = udp_idle_timer
        self.category = category
        self.icmptype = icmptype
        self.icmpcode = icmpcode

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f', name={self.name}, protocol={self.protocol}, tcp_portrange={self.tcp_portrange}'

    # Static Methods
    @staticmethod
    def _validate_and_get_portrange(prange):
        """ Return port range or list (as a string) of port ranges if port range is formatted as expected

        Args:
            prange (list): Port range. May be string or list of strings.  ex. '80'  or '100-300' or ['80', '100-300']

        Returns:
            String
        """
        if prange is None:
            return None
        else:
            # If range is provided as a single range in str format
            if isinstance(prange, str):
                # check that string is only numbers or number dash number
                if re.match(r'(^[\d]+$)|(^[\d]+-[\d]+$)', prange):
                    return prange
                else:
                    raise ValueError(f"portrange specified {prange} is not a valid.  Must be str of <digits> or "
                                     "<digits>-<digits>")

            # If a list of port ranges is provided in 'range' var
            elif isinstance(prange, list):
                range_list = ''
                for item in prange:
                    if isinstance(item, str):
                        # check that string is only numbers or number dash number
                        if re.match(r'(^[\d]+$)|(^[\d]+-[\d]+$)', item):
                            range_list += f' {item}'
                        else:
                            raise ValueError(
                                f"portrange specified: {item} is not a valid range.  Must be str of <digits> or "
                                "<digits>-<digits>")

                # Set self.<obj_type> with range_list values
                return range_list.lstrip()

    # Instance Property and Setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.name to name if name meets requirements

        Args:
            name (str): Name to set to self object

        Returns:
            None
        """
        if name is None:
            self._name = None

        else:
            if name.isspace():
                raise ValueError("'name', cannot be an empty string")

            if isinstance(name, str):
                if 1 <= len(name) <= 79:
                    self._name = name
                else:
                    raise ValueError("'name', must be type str() between 1 and 79 chars")
            else:
                raise ValueError("'name', must be type str()")

    @property
    def protocol(self):
        return self._protocol

    @protocol.setter
    def protocol(self, protocol):
        """ Set self.protocol to 'protocol' if valid 'protocol' provided

        Args:
            protocol (str): Type of protocol this object represents.  Options: 'tcp/udp/sctp', 'icmp', 'ip'

        Returns:
            None
        """
        if protocol is None:
            self._protocol = None
        else:
            if protocol.lower() in ['tcp', 'udp', 'sctp', 'tcp/udp/sctp']:
                self._protocol = 'TCP/UDP/SCTP'
            elif protocol.lower() == 'icmp':
                self._protocol = 'ICMP'
            elif protocol.lower() == 'ip':
                self._protocol = 'IP'
            else:
                raise ValueError("'protocol' must be type str() with value 'tcp/udp/sctp', 'icmp' or 'ip'")

    @property
    def tcp_portrange(self):
        return self.tcp_portrange

    @tcp_portrange.setter
    def tcp_portrange(self, prange):
        """ Set self.tcp_portrange to range returned from static method _validate_and_get_portrange()

        Args:
            prange (list):  Port range. May be string or list of strings.  ex. '80'  or '100-300' or ['80', '100-300']

        Returns:
            None
        """
        self._tcp_portrange = self._validate_and_get_portrange(prange)

    @property
    def udp_portrange(self):
        return self._udp_portrange

    @udp_portrange.setter
    def udp_portrange(self, prange):
        """ Set self.udp_portrange to range returned from static method _validate_and_get_portrange()

        Args:
            prange (list):  Port range. May be string or list of strings.  ex. '80'  or '100-300' or ['80', '100-300']

        Returns:
            None
        """
        self._udp_portrange = self._validate_and_get_portrange(prange)

    @property
    def sctp_portrange(self):
        return self._sctp_portrange

    @sctp_portrange.setter
    def sctp_portrange(self, prange):
        """ Set self.sctp_portrange to range returned from static method _validate_and_get_portrange()

        Args:
            prange (list):  Port range. May be string or list of strings.  ex. '80'  or '100-300' or ['80', '100-300']

        Returns:
            None
        """
        self._sctp_portrange = self._validate_and_get_portrange(prange)

    @property
    def protocol_number(self):
        return self._protocol_number

    @protocol_number.setter
    def protocol_number(self, protocol_number):
        """ Set self.protocol_number to protocol_number if protocol_number provided is valid. If not provided set None

        (Used when self.type is set to 'IP')

        Args:
            protocol_number (int): protocol number in range 0-254

        Returns:
            None
        """
        if protocol_number is None:
            self._protocol_number = None

        else:
            if isinstance(protocol_number, int):
                if 0 <= protocol_number <= 254:
                    self._protocol_number = protocol_number
                else:
                    raise ValueError("'protocol_number', when set, must be type int() between 1 and 254")
            else:
                raise ValueError("'protocol_number', when set, must be type int()")

    @property
    def visibility(self):
        return self.visibility

    @visibility.setter
    def visibility(self, visibility):
        """ Set self.visibility

        Args:
            visibility (str): object visibility in GUI.  ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if visibility is None:
            self._visibility = None

        else:
            if isinstance(visibility, str):
                if visibility == 'enable':
                    self._visibility = 'enable'
                elif visibility == 'disable':
                    self._visibility = 'disable'
                else:
                    raise ValueError("'visibility', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'visibility', when set, must be type str()")

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        """ Set self.comment to 'comment' if comment string within requirements

        Args:
            comment (str):  Comment for this policy object

        Returns:
            None
        """
        if comment is None:
            self._comment = None

        else:
            if isinstance(comment, str):
                if 1 <= len(comment) <= 255:
                    self._comment = comment
                else:
                    raise ValueError("'description', when set, must be type str() between 1 and 1,023 chars")
            else:
                raise ValueError("'description', when set, must be type str()")

    @property
    def session_ttl(self):
        return self._session_ttl

    @session_ttl.setter
    def session_ttl(self, session_ttl):
        """ Set self.session_ttl to session_ttl if session_ttl provided is valid

        Args:
            session_ttl (int): Session TTL for service object.  Acceptable range = 1-2764800

        Returns:
            None
        """
        if session_ttl is None:
            self._session_ttl = None

        else:
            if isinstance(session_ttl, int):
                if 300 <= session_ttl <= 2764800:
                    self._session_ttl = session_ttl
                else:
                    raise ValueError("'session_ttl', when set, must be type int between 300 and 2764800")
            else:
                raise ValueError("'session_ttl', when set, must be type int")

    @property
    def udp_idle_timer(self):
        return self._udp_idle_timer

    @udp_idle_timer.setter
    def udp_idle_timer(self, timer):
        """ set self.udp_idle_timer to timer if timer valid

        Args:
            timer (int): Udp idle timer for this service object.  Acceptable range = 0-86400

        Returns:
            None
        """
        if timer is None:
            self._udp_idle_timer = None

        else:
            if isinstance(timer, int):
                if 0 <= timer <= 864000:
                    self._udp_idle_timer = timer
                else:
                    raise ValueError("'udp_idle_timer', when set, must be type int between 0 and 86400")
            else:
                raise ValueError("'udp_idle_timer', when set, must be type int")

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, category):
        """ Set self.category if category provided is valid

        Args:
            category (str): Set FortiGate category for this service object.  String between 1 and 63 chars.

        Returns:
            None
        """
        if category is None:
            self._category = None

        else:
            if isinstance(category, str):
                if 1 <= len(category) <= 63:
                    self._category = category
                else:
                    raise ValueError("'category', when set, must be type str() between 1 and 63 chars")
            else:
                raise ValueError("'category', when set, must be type str()")

    @property
    def icmptype(self):
        return self.icmptype

    @icmptype.setter
    def icmptype(self, icmptype):
        """ Set self.icmptype to icmp_type if icmptype valid

        Args:
            icmptype (int): ICMP type.   Integer between 0 and 255.

        Returns:
            None
        """
        if icmptype is None:
            self._icmptype = None

        else:
            if isinstance(icmptype, int):
                if 0 <= icmptype <= 255:
                    self._icmptype = icmptype
                else:
                    raise ValueError("'icmptype', when set, must be type int between 0 and 255")
            else:
                raise ValueError("'icmptype', when set, must be type int")

    @property
    def icmpcode(self):
        return self._icmpcode

    @icmpcode.setter
    def icmpcode(self, icmpcode):
        """ Set self.icmpcode to icmpcode if icmp_type valid

        Args:
            icmpcode (int): ICMP code.   Integer between 0 and 255.

        Returns:
            None
        """
        if icmpcode is None:
            self._icmpcode = None

        else:
            if isinstance(icmpcode, int):
                if 0 <= icmpcode <= 255:
                    self._icmpcode = icmpcode
                else:
                    raise ValueError("'icmpcode', when set, must be type int between 0 and 255")
            else:
                raise ValueError("'icmpcode', when set, must be type int")



import ipaddress

from fgobjlib import FgObject


class FgRouteIPv4(FgObject):
    """FgRouteIPv4 class represents FortiGate Firewall static route object and provides methods for validating
    parameters and generating both cli and api configuration data for use in external configuration applications

    Attributes:
        routeid (str): ID of this object.
        dst (str): Destination Network, an IPv4 Network
        device (str): Destination interface
        gateway (str): Next-hop gateway, an IPv4 Address
        distance (int): Route distance
        priority (int): Route priority
        weight (int): Route weight
        comment (str): Route comment
        blackhole (str): Blackhole route ('enable', 'disable', or None=inherit)
        vrf (int): vrf ID for route
        vdom (str): vdom for this route
    """

    def __init__(self, routeid: int = None, dst: str = None, device: str = None, gateway: str = None,
                 distance: int = None, priority: int = None, weight: int = None, comment: str = None,
                 blackhole: str = None, vrf: int = None, vdom: str = None):
        """
        Args:
            routeid (int): ID for route object.  If not set, defaults to 0.
            dst (str): Destination network for route.  Must be valid IPv4 network/mask.  If no mask, defaults to /32.
            device (str): Destination interface for route.
            gateway (str): Next-hop gateway for route.  Must be valid IPv4 address.
            distance (int): Distance for route.
            priority (int): Priority for route.
            weight (int): Weight for route.
            comment (str): Comment for route
            blackhole (str): Enable or disable blackhole route.   ('enable', 'disable', or None=inherit)
            vrf (int): VRF number to set for route
            vdom (str): VDOM, if applicable, for route
        """

        # Initialize the parent class - we do set this here, because the subclass will first verify obj_id
        # is acceptable for this class type in the above attribute set functions
        super().__init__(api='cmdb', api_path='router', api_name='static', cli_path="config router static",
                         obj_id=routeid, vdom=vdom)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'routeid': 'seq-num', 'dst': 'dst', 'device': 'device', 'gateway': 'gateway',
                            'distance': 'distance', 'priority': 'priority', 'weight': 'weight', 'comment': 'comments',
                            'blackhole': 'blackhole', 'vrf': 'vrf'}

        self._cli_ignore_attrs = ['routeid']

        # Set instance attributes #
        self.routeid = routeid
        self.dst = dst
        self.device = device
        self.gateway = gateway
        self.distance = distance
        self.priority = priority
        self.weight = weight
        self.comment = comment
        self.blackhole = blackhole
        self.vrf = vrf

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f", routeid={self.routeid}, dst={self.dst}, device={self.device}, gateway={self.gateway}," \
                            f"distance={self.distance}, priority={self.priority}, weight={self.weight}, " \
                            f"comment={self.comment}, blackhole={self.blackhole}, vrf={self.vrf}, vdom={self.vdom}"

    # Class Methods
    @classmethod
    def blackhole_route(cls, routeid: int = 0, dst: str = None, vdom: str = None, distance: int = None,
                        priority: int = None, weight: int = None, comment: str = None, vrf: int = None):
        """ Class Method to streamline config for blackhole  routes

        Args:
            routeid (int): ID of route, if not set defaults to 0
            dst (str): Destination network for route
            vdom (str): VDOM, if applicable for route
            distance (int): Set route distance
            priority (int): Set route priority
            weight (int): Set route weight
            comment (str): Set route comment
            vrf (int): Set route VRF

        Returns:
            Class Instance
        """

        device = None
        gateway = None
        blackhole = 'enable'

        obj = cls(routeid, dst, device, gateway, distance, priority, weight, comment, blackhole, vrf, vdom)
        return obj

    # Instance properties and setters
    @property
    def routeid(self):
        return self._routeid

    @routeid.setter
    def routeid(self, routeid):
        """ Set self.routeid to routeid if routeid is provided and valid, else set self.routeid to 0

        Args:
            routeid (int): Id for route.  If routeid = None, self.routeid set to 0

        Returns:
            None
        """
        if routeid is None:
            self._routeid = 0

        else:
            if isinstance(routeid, int):
                if 0 <= routeid <= 4294967295:
                    self._routeid = routeid
                else:
                    raise ValueError("'routeid' must be type int() between 0 and 4294967295")
            else:
                raise ValueError("'routeid', must type int()")

    @property
    def dst(self):
        return self._dst

    @dst.setter
    def dst(self, dst):
        """ Set self.dst to dst if dst is valid ipv4 network/mask.  If no mask is set, mask defaults to 32.

        Args:
            dst (str): A valid IPv4 network/mask

        Returns:
            None
        """
        if dst is None:
            self._dst = None

        else:
            if isinstance(dst, str):
                try:
                    self._dst = str(ipaddress.ip_network(dst))
                except ValueError:
                    raise ValueError("'dst' must be type str() with value containing a valid ipv4 network and mask")
            else:
                raise ValueError("'dst' must be type: str()")

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device):
        """ Set self.device to device if device contains valid values

        Args:
            device (str): Route destination device

        Returns:
            None
        """
        if device is None:
            self._device = None

        else:
            if isinstance(device, str) and 1 <= len(device) <= 35:
                self._device = device
            else:
                raise ValueError("'device' must be type str() between 1 and 35 chars", device)

    @property
    def gateway(self):
        return self._gateway

    @gateway.setter
    def gateway(self, gateway):
        """ Set self.gateway to gateway if gateway is valid ipv4 address

        Args:
            gateway (str): Next-hop gateway.  Must be valid ipv4 address.

        Returns:
            None
        """
        if gateway is None:
            self._gateway = None

        else:
            if isinstance(gateway, str):
                try:
                    self._gateway = str(ipaddress.ip_address(gateway))
                except ValueError:
                    raise ValueError("'gateway', when set, must be type str() containing a valid ipv4 address address")
            else:
                raise ValueError("'gateway', when set must be type: str() containing a valid ipv4 address")

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, distance):
        """ Set self.distance to distance if distance valid

        Args:
            distance (int): Route distance

        Returns:
            None
        """
        if distance is None:
            self._distance = None

        else:
            if isinstance(distance, int) and 1 <= distance <= 255:
                self._distance = distance
            else:
                raise ValueError("'distance', when set, must be type int() with value between 1 and 255")

    @property
    def weight(self):
        return self._distance

    @weight.setter
    def weight(self, weight):
        """ Set self.weight to weight if weight valid

        Args:
            weight (int): Route weight

        Returns:

        """
        if weight is None:
            self._weight = None

        else:
            if isinstance(weight, int) and 1 <= weight <= 255:
                self._weight = weight
            else:
                raise ValueError("'weight', when set, must be type int() with value between 1 and 255")

    @property
    def priority(self):
        return self._priority

    @priority.setter
    def priority(self, priority):
        """ Set self.priority to priority if priority valid

        Args:
            priority (int): Route priority

        Returns:
            None
        """
        if priority is None:
            self._priority = None

        else:
            if isinstance(priority, int) and 0 <= priority <= 4294967295:
                self._priority = priority
            else:
                raise ValueError("'priority', when set, must be type int() with value between 0 and 4294967295")

    @property
    def vrf(self):
        return self._vrf

    @vrf.setter
    def vrf(self, vrf):
        """ Set self.vrf to vrf if vrf is valid

        Args:
            vrf (int): Route VRF

        Returns:
            None
        """
        if vrf is None:
            self._vrf = None

        else:
            if isinstance(vrf, int) and 0 <= vrf <= 31:
                self._vrf = vrf
            else:
                raise ValueError("'vrf', when set, must be type int() with value between 0 and 31")

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        """ Set self.comment to comment if comment valid

        Args:
            comment: Route comment

        Returns:
            None
        """
        if comment is None:
            self._comment = None

        else:
            if isinstance(comment, str) and 1 <= len(comment) <= 255:
                self._comment = comment
            else:
                raise Exception("'comment', when set, must be type str() between 1 and 255 chars")

    @property
    def blackhole(self):
        return self._blackhole

    @blackhole.setter
    def blackhole(self, blackhole):
        """ Set self.blackhole to True, False if blackhole = True or False, else set to None

        Args:
            blackhole: Set blackhole True=enable, False=disable, None=inherit

        Returns:
            None
        """
        if blackhole is None:
            self._blackhole = None

        else:
            if isinstance(blackhole, bool):
                if blackhole == 'enable':
                    self._blackhole = 'enable'
                elif blackhole == 'disable':
                    self._blackhole = 'diable'
                else:
                    raise ValueError("'blackhole', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'blackhole', when set, must be type str()")



import ipaddress
from typing import Union

from fgobjlib import FgObject


class FgIpsecP2Interface(FgObject):
    """
    FgIpsecP2Interface class represents FortiGate Firewall ipsec phase2 interface object and provides methods for
    validating parameters and generating both cli and api configuration data for use in external configuration
    applications and ftntlib

    Currently supports dynamic or static VPN using psk authentication. No support yet for advpn or mode-cfg or pki

    Attributes:
        name (str): Name of phase2-interface
        phase1name (str): Name of phase1-interface to bind to
        proposal (str): phase2 proposal(s)
        pfs (str): perfect forward secrecy,  should be either 'enable' or 'disable'
        dhgrp (str): phase2 dhgrp(s)
        keepalive (str): keepalive, should be either 'enable' or 'disable'
        replay (str): replay protection
        comment (str): phase2 comment
        auto_negotiate (str): auto-negotiation, should be either 'enable' or 'disable'
        vdom (str): associated VDOM
        src_subnet (str):  source selector, for selectors type subnet
        dst_subnet (str): destination selector, for selectors type subnet
    """

    def __init__(self, name: str = None, phase1name: str = None, proposal: list = None, pfs: str = None,
                 dhgrp: Union[str, list] = None, keepalive: str = None, replay: str = None, comment: str = None,
                 auto_negotiate: str = None, vdom: str = None, src_subnet: str = None, dst_subnet: str = None):
        """
        Args:
            name (str): Name of phase2-interface
            phase1name (str): Name of phase1-interface to bind to
            proposal (str): phase2 proposal(s)
            pfs (str): perfect forward secrecy.  Allowed values are 'enable' or 'disable'
            dhgrp (str): phase2 dhgrp(s)
            keepalive (str): keepalive.  Allowed values are 'enable' or 'disable'
            replay (str): replay protection (True=enabled, False=disable, None=inherit)
            comment (str): phase2 comment
            auto_negotiate (str): Auto-negotiation. Allowed values are 'enable' or 'disable'
            vdom (str): associated VDOM
            src_subnet (str):  source selector, for selectors type subnet
            dst_subnet (str): destination selector, for selectors type subnet
        """

        # Initialize the parent class
        super().__init__(api='cmdb', api_path='vpn.ipsec', api_name='phase2-interface',
                         cli_path="config vpn ipsec phase2-interface", obj_id=name, vdom=vdom)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name', 'phase1name': 'phase1name', 'proposal': 'proposal',
                            'comment': 'comments', 'keepalive': 'keepalive', 'dhgrp': 'dhgrp', 'pfs': 'pfs',
                            'replay': 'replay', 'auto_negotiate': 'auto-negotiate', 'src_subnet': 'src-subnet',
                            'dst_subnet': 'dst-subnet'}

        self._cli_ignore_attrs = ['name']

        # Set instance attributes #
        self.name = name
        self.phase1name = phase1name
        self.proposal = proposal
        self.comment = comment
        self.keepalive = keepalive
        self.dhgrp = dhgrp
        self.pfs = pfs
        self.replay = replay
        self.auto_negotiate = auto_negotiate
        self.src_subnet = src_subnet
        self.dst_subnet = dst_subnet

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f', name={self.name}, phase1name={self.phase1name}, proposal={self.proposal}, ' \
                            f'comment={self.comment}, keepalive={self.keepalive}, dhgrp={self.dhgrp}, ' \
                            f'pfs={self.pfs}, replay={self.replay}, auto_negotiate={self.auto_negotiate}, ' \
                            f'src_subnet={self.src_subnet}, dst_subnet={self.dst_subnet}'

    # Instance properties and setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.name to name if name valid

        Args:
            name (str): Name for phase2-interface

        Returns:
            None
        """
        if name is None:
            self._name = None
        else:
            if name.isspace():
                raise ValueError("'name', cannot be an empty")
            if isinstance(name, str):
                if len(name) <= 35:
                    self._name = name
                else:
                    raise ValueError("'name', must be type str() between 1 and 35 chars")
            else:
                raise ValueError("'name', must be type str()")

    @property
    def phase1name(self):
        return self._phase1name

    @phase1name.setter
    def phase1name(self, phase1name):
        """ Set self.phase1name to phase1name if phase1name valid

        Args:
            phase1name (str): phase1name to bind to

        Returns:
            None
        """
        if phase1name is None:
            self._phase1name = None
        else:
            if phase1name.isspace():
                raise ValueError("'phase1name', cannot be an empty string")
            if isinstance(phase1name, str):
                if len(phase1name) <= 35:
                    self._phase1name = phase1name
                else:
                    raise ValueError("'phase1name', must be type str between 1 and 35 chars")
            else:
                raise ValueError("'phase1name', must be type str()")

    @property
    def proposal(self):
        return self._proposal

    @proposal.setter
    def proposal(self, proposal):
        """ Set self.proposal to proposal if proposal contains valid FG proposals

        Args:
            proposal: phase2 proposal.  May be string or list of strings.  i.e. 'des-md5' or ['des-md5', '3des-md5']

        Returns:
            None
        """
        valid_proposals = ['des-md5', 'des-sha', 'des-sha256', 'des-sha384', 'des-sha512', '3des-md5', '3des-sha1',
                           '3des-sha256', '3des-sha384', '3des-sha512', 'aes128-md5', 'aes128-sha1', 'aes128-sha256',
                           'aes128-sha384', 'aes128-sha512', 'aes192-md5', 'aes192-sha1', 'aes192-sha256',
                           'aes192-sha384', 'aes192-sha512', 'aes256-md5', 'aes256-sha1', 'aes256-sha256',
                           'aes256-sha384', 'aes256-sha512', 'aria128-md5', 'aria128-sha1', 'aria128-sha256',
                           'aria128-sha384', 'aria128-sha512', 'aria192-md5', 'aria192-sha1', 'aria192-sha256',
                           'aria192-sha384', 'aria192-sha512', 'aria256-md5', 'aria256-sha1', 'aria256-sha256',
                           'aria256-sha384', 'aria256-sha512', 'seed-md5', 'seed-sha1', 'seed-sha256', 'seed-sha384',
                           'seed-sha512', 'chacha20poly1305', 'null-md5', 'null-sha1', 'null-sha256', 'null-sha384',
                           'null-sha512', 'des-null', '3des-null', 'aes128-null', 'aes192-null', 'aes256-null',
                           'aria128-null', 'seed-null']

        if proposal is None:
            self._proposal = None

        else:
            proposal_items = ''

            # IF a single object was passed as a string, append it to intf_list else iterate the list and pull
            # out the strings of interfaces and append each to intf_list
            if isinstance(proposal, str):

                # compare proposal to valid_proposals list
                if proposal in valid_proposals:
                    proposal_items += f"{proposal}"
                else:
                    raise ValueError("'proposal' value provided is not a valid fortigate phase1 proposal option")

            elif isinstance(proposal, list):
                for item in proposal:
                    if isinstance(item, str):

                        # compare proposal to valid proposals list
                        if item in valid_proposals:
                            proposal_items += f" {item}"
                        else:
                            raise ValueError("'proposal' value provided not a valid fortigate phase1 proposal option")
            else:
                raise ValueError("'proposal' must be type str() with single proposal referenced or type list() for "
                                 "multiple proposal references")

            self._proposal = proposal_items

    @property
    def dhgrp(self):
        return self._dhgrp

    @dhgrp.setter
    def dhgrp(self, dhgrp):
        """  Set self.dhgrp to string containing values dhgrp if dhgrp contains valid FortiGate proposals

        The dhgrp may be passed in as a str, or a list.  If the dhgrp(s) passed in are valid we'll add each
        of those to a space separated values string and set in self.dhgrp

        Args:
            dhgrp (list): list of valid fortigate dhgrps

        Returns:
            None
        """
        if dhgrp is None:
            self._dhgrp = None

        else:
            dhgrp_items = ''
            valid_dhgrps = [1, 2, 5, 14, 15, 16, 17, 18, 19, 20, 21, 27, 28, 30, 31, 32]

            # IF a single object was passed as a string, append it to list else iterate the list and pull
            # out the dhgrps and add to list to be set as self object
            if isinstance(dhgrp, int):
                # compare proposal to valid_list
                if dhgrp in valid_dhgrps:
                    dhgrp_items += f"{dhgrp} "
                else:
                    raise ValueError(f"'dhgrp' value provided is not a valid fortigate dhgrp option")

            elif isinstance(dhgrp, list):
                for item in dhgrp:
                    if isinstance(item, int):

                        # compare proposal to valid proposals list
                        if item in valid_dhgrps:
                            dhgrp_items += f"{item} "
                        else:
                            raise ValueError(f"At least one 'dhgrp' provided is not a valid fortigate phase1 proposal")
            else:
                raise ValueError("dhgrp must be provided as type int() or list() of int()")

            self._dhgrp = dhgrp_items

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        """ Set self.comment to comment if comment valid

        Args:
            comment (str): phase2-interface comment

        Returns:
            None
        """
        if comment is None:
            self._comment = None
        else:
            if isinstance(comment, str):
                if 1 <= len(comment) <= 1023:
                    self._comment = comment
                else:
                    raise ValueError("'description', when set, must be type str() between 1 and 1,023 chars")
            else:
                raise ValueError("'description', when set, must be type str()")

    @property
    def keepalive(self):
        return self._keepalive

    @keepalive.setter
    def keepalive(self, keepalive):
        """ Set self.keepalive to keepalive if keepalive valid

        Args:
            keepalive (str):  keepalive, allowed values: 'enable' or 'disable'

        Returns:
            None
        """
        if keepalive is None:
            self._keepalive = None
        else:
            if isinstance(keepalive, str):
                if keepalive == 'enable':
                    self._keepalive = 'enable'
                elif keepalive == 'disable':
                    self._keepalive = 'disable'
                else:
                    raise ValueError("'keepalive' when set must be type str() with value 'enable' or 'disable'")

            else:
                raise ValueError("'keepalive', when set, must be type str()")

    @property
    def pfs(self):
        return self._pfs

    @pfs.setter
    def pfs(self, pfs):
        """ set self.pfs to enable, disable or None based on pfs value

        Args:
            pfs (str): pfs, allowed values: 'enable'  or 'disable'

        Returns:
            None
        """
        if pfs is None:
            self._pfs = None
        else:
            if isinstance(pfs, str):
                if pfs == 'enable':
                    self._pfs = 'enable'
                elif pfs == 'disable':
                    self._pfs = 'disable'
                else:
                    raise ValueError("'pfs', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'pfs', when set, must be type str()")

    @property
    def replay(self):
        return self._replay

    @replay.setter
    def replay(self, replay):
        """ Set self.replay to enable, disable or None

        Args:
            replay (str): replay protection, allowed values: 'enable' or 'disable'

        Returns:
            None
        """
        if replay is None:
            self._replay = None
        else:
            if isinstance(replay, str):
                if replay == 'enable':
                    self._replay = 'enable'
                elif replay == 'disable':
                    self._replay = 'disable'
                else:
                    raise ValueError("'replay', when set must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'pfs', when set, must type str()")

    @property
    def auto_negotiate(self):
        return self.auto_negotiate

    @auto_negotiate.setter
    def auto_negotiate(self, auto_negotiate):
        """ Set self.auto_negotiate to enable, disable or None

        Args:
            auto_negotiate (str): auto-negotiation, allowed values: 'enable' or 'disable'

        Returns:
        None
        """
        if auto_negotiate is None:
            self._auto_negotiate = None
        else:
            if isinstance(auto_negotiate, str):
                if auto_negotiate == 'enable':
                    self._auto_negotiate = 'enable'
                elif auto_negotiate == 'disable':
                    self._auto_negotiate = 'disable'
                else:
                    raise ValueError("'auto_negotiate', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'pfs', when set, must type str()")

    @property
    def src_subnet(self):
        return self._src_subnet

    @src_subnet.setter
    def src_subnet(self, src_subnet):
        """ Set self.src_subnet if src_subnet valid

        Args:
            src_subnet: Source Selector when selector type set to subnet.  Must be valid ipv4 network/mask.

        Returns:
            None
        """
        if src_subnet is None:
            self._src_subnet = None
        else:
            try:
                ipaddress.ip_network(src_subnet)
            except ValueError:
                raise ValueError("'src_subnet', when set, must be a valid ipv4 or ipv6 address")
            else:
                self._src_subnet = src_subnet

    @property
    def dst_subnet(self):
        return self._dst_subnet

    @dst_subnet.setter
    def dst_subnet(self, dst_subnet):
        """ Set self.dst_subnet if dst_subnet valid

        Args:
            dst_subnet: Destination Selector when selector type set to subnet.  Must be valid ipv4 network/mask.

        Returns:
            None
        """
        if dst_subnet is None:
            self._dst_subnet = None
        else:
            try:
                ipaddress.ip_network(dst_subnet)
            except ValueError:
                raise ValueError("'dst_subnet', when set, must be type str() with value a valid ipv4 or ipv6 address")
            else:
                self._dst_subnet = dst_subnet



import ipaddress

from fgobjlib import FgObject


class FgInterfaceIpv4(FgObject):
    """
    FgInterface class represents FortiGate Firewall interface object and provides methods for validating parameters
    and generating both cli and api configuration data for use in external configuration applications

    Currently supports interface types of "standard" i.e. ethernet/physical or vlan

    Attributes:
        name (str): Name of interface
        ip (str): IP address of interface
        mode (str): Interface mode, i.e. static or dhcp
        intf_type (str): Interface type, i.e. physical/virtual (default) or vlan
        vdom (str): Vdom interface configured for
        vrf (int): VRF interface is configured for
        allowaccess (str):  Interface allowaccess policy.  i.e. "ping http, https snmp, etc"
        role (str):  Interface role type
        vlanid (str): Interface vlanid
        phys_intf (str): Interfaces physical intf attachment.  (if intf_type is vlan)
        device_ident (str): device-identification ('enable', 'disable', or None=inherit)
        alias (str): Interface alias
        description (str): Interface description
    """

    def __init__(self, name: str = None, ip: str = None, mode: str = None, intf_type: str = None, vdom: str = None,
                 vrf: int = None, allowaccess: str = None, role: str = None, vlanid: int = None, phys_intf: str = None,
                 device_ident: str = None, alias: str = None, description: str = None, is_global: bool = None):
        """
        Args:
            name (str): Name of interface
            ip (str): IP address of interface
            mode (str): Interface mode, i.e. static or dhcp
            intf_type (str): Interface type, i.e. physical/virtual (default) or vlan
            vdom (str): Vdom interface configured for
            vrf (int): VRF interface is configured for
            allowaccess (str):  Interface allowaccess policy.  i.e. "ping http, https snmp, etc"
            role (str):  Interface role type
            vlanid (str): Interface vlanid
            phys_intf (str): Interfaces physical intf attachment.  (if intf_type is vlan)
            device_ident (str): device-identification  ('enable', 'disable', or None=inherit)
            alias (str): Interface alias
            description (str): Interface description
        """

        # Initialize the parent class
        super().__init__(api='cmdb', api_path='system', api_name='interface', cli_path="config system interface",
                         obj_id=name, vdom=vdom)

        # Set parent class attributes #
        # Map instance attribute names to fg attribute names
        self._data_attrs = {'name': 'name', 'ip': 'ip', 'vdom': 'vdom', 'intf_type': 'type', 'vrf': 'vrf',
                            'allowaccess': 'allowaccess', 'role': 'role', 'vlanid': 'vlanid',
                            'phys_intf': 'interface', 'device_ident': 'device-identification',
                            'alias': 'alias', 'description': 'description'}

        # Attributes to ignore for cli config
        self._cli_ignore_attrs = ['name']

        # For objects types that can be configured from global or vdom context,
        # set is_global = True if need to config from global context instead of vdom
        self.is_global = is_global

        # Set instance attributes
        self.name = name
        self.ip = ip
        self.mode = mode
        self.intf_type = intf_type
        self.vrf = vrf
        self.allowaccess = allowaccess
        self.role = role
        self.vlanid = vlanid
        self.phys_intf = phys_intf
        self.device_ident = device_ident
        self.alias = alias
        self.description = description

        # Update the parent defined obj_to_str attribute with this objects str representation
        self._obj_to_str += f', name={self.name}, ip={self.ip}, mode={self.mode}, intf_type={self.intf_type}, ' \
                            f'vrf={self.vrf}, allowaccess={self.allowaccess}, role={self.role}, ' \
                            f'vlanid={self.vlanid}, phys_intf={self.phys_intf}.device_ident={self.device_ident}, ' \
                            f'alias={self.alias}, description={self.description}, vdom={self.vdom}'

    # Class Methods
    @classmethod
    def standard_intf(cls, name: str, ip: str = None, mode: str = None, vdom: str = None, vrf: int = None,
                      allowaccess: str = None, role: str = None, device_ident: bool = None, alias: str = None,
                      description: str = None):
        """ Class method to streamline instantiating instance of type "standard" aka ethernet/virtual/physical

        Args:
            name (str): Name of interface
            ip (str): IP address of interface
            mode (str): Interface mode, i.e. static or dhcp
            vdom (str): Vdom interface configured for
            vrf (int): VRF interface is configured for
            allowaccess (str):  Interface allowaccess policy.  i.e. "ping http, https snmp, etc"
            role (str):  Interface role type
            device_ident (str): device-identification  ('enable', 'disable', or None=inherit)
            alias (str): Interface alias
            description (str): Interface description

        Returns:
            Class Instance
        """

        intf_type = None
        vlanid = None
        phys_intf = None

        obj = cls(name=name, ip=ip, mode=mode, intf_type=intf_type, vdom=vdom, vrf=vrf, allowaccess=allowaccess,
                  role=role, vlanid=vlanid, phys_intf=phys_intf, device_ident=device_ident, alias=alias,
                  description=description)

        return obj

    @classmethod
    def vlan_intf(cls, name: str, vlanid: int, phys_intf: str, ip: str = None, mode: str = 'static', vdom: str = None,
                  vrf: int = None, allowaccess: str = None, role: str = None, device_ident: bool = None,
                  alias: str = None, description: str = None):
        """ Class method to streamline instantiating instance of type vlan

        Args:
            name (str): Name of interface
            vlanid (int): interface vlanid
            phys_intf (str): name of parent interface for vlan attachment
            ip (str): IP address of interface
            mode (str): Interface mode, i.e. static or dhcp
            vdom (str): Vdom interface configured for
            vrf (int): VRF interface is configured for
            allowaccess (str):  Interface allowaccess policy.  i.e. "ping http, https snmp, etc"
            role (str):  Interface role type
            device_ident (str): device-identification  ('enable', 'disable', or None=inherit)
            alias (str): Interface alias
            description (str): Interface description

        Returns:
            Class Instance
        """

        if not vlanid:
            raise Exception("'vlan_intf()', requires to provide a 'vlanid'. vlanid should be type: int, between 1 and "
                            "4960")

        if not phys_intf:
            raise Exception("'vlan_intf', requires to provide 'phys_intf', phys_intf should be type: str between 1 and "
                            "15 chars")

        intf_type = 'vlan'

        obj = cls(name=name, ip=ip, mode=mode, intf_type=intf_type, vdom=vdom, vrf=vrf, allowaccess=allowaccess,
                  role=role, vlanid=vlanid, phys_intf=phys_intf, device_ident=device_ident, alias=alias,
                  description=description)

        return obj

    # Instance Properties and Setters
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        """ Set self.name to name if name meets requirements

        Args:
            name (str): Interface name to set.

        Returns:
            None
        """
        if name is None:
            self._name = None

        else:
            if isinstance(name, str):
                if name.isspace():
                    raise ValueError("'name', cannot be an empty string")
                if len(name) < 15:
                    self._name = name
                else:
                    raise ValueError("'name', must be less than 15 chars")
            else:
                raise ValueError("'name', must be a string")

    @property
    def ip(self):
        return self._ip

    @ip.setter
    def ip(self, ip):
        """ Set self.ip as IP address for interface if IP is valid ipv4 address.  Otherwise raise Exception.

        Args:
            ip: Interface IPv4 address to set.  Must be valid IP/netmask.  IP without netmask = /32

        Returns:
            None
        """
        if ip is None:
            self._ip = None

        else:
            try:
                self._ip = str(ipaddress.ip_interface(ip))
            except ValueError:
                raise ValueError("'ip' must be type str() as valid ipv4 address")

    @property
    def intf_type(self):
        return self._intf_type

    @intf_type.setter
    def intf_type(self, intf_type):
        """ Set self.intf_type if intf type matches allowed types

        Args:
            intf_type (str): Interface type: 'vlan', 'standard', 'loopback', None

        Returns:
            None
        """
        if intf_type is None:
            self._intf_type = None

        else:
            if isinstance(intf_type, str):
                if intf_type.lower() == 'vlan':
                    self. _intf_type = 'vlan'
                elif intf_type.lower() == 'standard':
                    self._intf_type = None
                elif intf_type.lower() == 'loopback':
                    self._intf_type = 'loopback'
                else:
                    raise ValueError("'intf_type' when set, must be type str() with value 'vlan', 'standard' or "
                                     "'loopback")
            else:
                raise ValueError("'intf_type', when set, must be type str()")

    @property
    def role(self):
        return self._role

    @role.setter
    def role(self, role):
        """ Set self.role if role matches allowed role types.

        Args:
            role (str): Interface role.  Allowed types: 'wan', 'lan', 'dmz', 'undefined', None (use existing setting)

        Returns:

        """
        if role is None:
            self._role = None

        else:
            if isinstance(role, str):
                if role.lower() == 'wan':
                    self._role = 'wan'
                elif role.lower() == 'lan':
                    self._role = 'lan'
                elif role.lower() == 'dmz':
                    self._role = 'dmz'
                else:
                    self._role = 'undefined'
            else:
                raise ValueError("'role', when set, must be type str()")

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        """ Set self.mode to mode if mode set with valid argument

        Args:
            mode (str): Interface Mode to set.   May be: 'dhcp', 'static', None (None uses existing intf setting)

        Returns:
            None
        """
        if mode is None:
            self._mode = 'dhcp'

        else:
            if isinstance(mode, str):
                if mode.lower() == 'dhcp':
                    self._mode = 'dhcp'
                elif mode.lower() == 'static':
                    self._mode = 'static'
                else:
                    raise ValueError("'mode', when set, must be type str() with value 'dhcp' or 'static'")
            else:
                raise ValueError("'mode', when set, must be type str()")

    @property
    def allowaccess(self):
        return self._allowaccess

    @allowaccess.setter
    def allowaccess(self, allowaccess):
        """ Set self.allowaccess  to allowaccess if allowaccess contains allowed params

        Args:
            allowaccess (str): Interface mgmt access setting. Single string with spaces to separate multiple values

        Returns:
            None
        """
        if allowaccess is None:
            self._allowaccess = None

        else:
            if isinstance(allowaccess, str):
                for service in list(allowaccess.split(" ")):
                    if service.lower() in ['ping', 'http', 'https', 'snmp', 'ssh', 'telnet', 'fgfm', 'radius=acct',
                                           'probe-response', 'capwap', 'ftm']:
                        continue
                    else:
                        raise ValueError("'allowaccess' has unrecognized services defined")
            else:
                raise ValueError("'allowaccess', when set, must be type str()")

            self._allowaccess = allowaccess

    @property
    def vlanid(self):
        return self._vlanid

    @vlanid.setter
    def vlanid(self, vlanid):
        """  Set self.vlanid if vlanid is valid

        Args:
            vlanid: Interface vlanid to set.  Must be between 1 and 4096

        Returns:
            None
        """
        if vlanid is None:
            self._vlanid = None

        else:
            if isinstance(vlanid, int):
                if 1 <= vlanid <= 4096:
                    self._vlanid = vlanid
                else:
                    raise ValueError("'vlanid', when set, must be type int() between 1 and 4096")
            else:
                raise ValueError("vlanid, when set, must be type int()")

    @property
    def phys_intf(self):
        return self._phys_intf

    @phys_intf.setter
    def phys_intf(self, phys_intf):
        """ Set self.phys_intf to phys_intf if phys_intf valid

        Args:
            phys_intf (str): Parent interface for virtual interface type, such as vlan to set.

        Returns:
            None
        """
        if phys_intf is None:
            self._phys_intf = None

        else:
            if isinstance(phys_intf, str):
                if phys_intf.isspace():
                    raise ValueError("'phys_intf' cannot be an empty string")

                if 1 <= len(phys_intf) <= 31:
                    self._phys_intf = phys_intf
                else:
                    raise ValueError("'phys_intf', when set, must be type str() between 1 and 31 chars")
            else:
                raise ValueError("'phys_intf', when set, must be type str()")

    @property
    def vrf(self):
        return self._vrf

    @vrf.setter
    def vrf(self, vrf):
        """ Set self.vrf if vrf valid

        Args:
            vrf (int): Interface VRF to set

        Returns:
            None
        """
        if vrf is None:
            self._vrf = None

        else:
            if isinstance(vrf, int):
                if 0 <= vrf <= 31:
                    self._vrf = vrf
                else:
                    raise ValueError("'vrf', when set, must be type int() between 0 and 31")
            else:
                raise ValueError("'vrf', when set, must be type int() between 0 and 31")

    @property
    def device_ident(self):
        return self._device_ident

    @device_ident.setter
    def device_ident(self, device_ident):
        """ Set self.device_ident if device_ident valid

        Args:
            device_ident (str): device-identification  ('enable', 'disable', or None=inherit)

        Returns:
            None
        """
        if device_ident is None:
            self._device_ident = False

        else:
            if isinstance(device_ident, str):
                if device_ident == 'enable':
                    self._device_ident = 'enable'
                elif device_ident == 'disable':
                    self._device_ident = 'disable'
                else:
                    raise ValueError("'device_ident', when set, must be type str() with value 'enable' or 'disable'")
            else:
                raise ValueError("'device_ident', when set, must be type str()")

    @property
    def alias(self):
        return self._alias

    @alias.setter
    def alias(self, alias):
        """ Set self.alias to alias if alias valid

        Args:
            alias: Interface Alias to set

        Returns:

        """
        if alias is None:
            self._alias = None

        else:
            if isinstance(alias, str):
                if 1 <= len(alias) <= 25:
                    self._alias = alias
                else:
                    raise ValueError("'alias', when set, must be type str() between 1 and 25 chars")
            else:
                raise ValueError("'alias', when set, must be type str()")

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        """ Set self.description to description if description is valid

        Args:
            description: Interface description to set.

        Returns:
            None
        """
        if description is None:
            self._description = None

        else:
            if isinstance(description, str):
                if 1 <= len(description) <= 255:
                    self._description = description
                else:
                    raise Exception("'description', when set, must be type str() between 1 and 255 chars")
            else:
                raise Exception("'description', when set, must be type str()")



from abc import ABC

class FgObject(ABC):
    """FgObject class represents basic methods and attributes used commonly across most, if not all, child class objects

    This is an abstract class and as such cannot be instantiated directly.

    FgObject class provides attributes and methods that are used by the specific FortiGate object child classes.  Common
    attributes such as VDOM and object ID are contained and validated in this parent class.  API and CLI path attributes
    are also represented here.  Various methods used by most of the child classes for returning CLI or API configuration
    data are also within this FgObject class. This class is typically not called directly by an application, rather it
    is extended by the specific FortiGate object child classes such as FgFwPolicy.

    Attributes:
        API (str): string representing the FortiGate API to be called.  (this will almost always be 'cmdb')
        API_PATH (str): string representing the api path associated with this object. (ex. system, router, firewall, etc..)
        API_NAME (str): string representing the object type within the api_path to add/update/delete/get.
        API_MKEY: updated with object_id for update/delete/get operations
        CLI_PATH (str):  string with base FG cli context configurations for this object (ex: "config system interface")
        obj_id: this may be string, integer or other type that defines the object ID for use in api and cli config.
        vdom: defines the vdom this object is to be configured/updated/deleted/got from.
        vdom_enabled: used for objects which "may" be configured from global context to determine correct cli config path
        is_global (bool): Set if the object should be configured from global context only
"""

    def __init__(self, api: str = None, api_path: str = None, api_name: str = None,  cli_path = None,
                 obj_id = None, vdom: str = None):
        """
        Args:
            api (str): set the FortiGate API to use. (default None) - Required if using API config methods.
            api_path (str): optional - set the FortiGate API path. (default None) - Required if using API config methods.
            api_name (str): optional - set the FortiGate API name.  (default None) - Required if using API config methods
            cli_path (str): optional - set the FortiGate CLI configuration path . (default None) -
            Required if using CLI config methods
            obj_id: required - set the object id, this may be str or int.
            vdom (str): optional - set the objects vdom, if FG vdoms enabled and object type is configured in vdom context
        """

        # Set the API path
        self.API = api
        self.API_PATH = api_path
        self.API_NAME = api_name
        self.API_MKEY = None

        # Set the CLI path
        self.CLI_PATH = cli_path

        # Set attrs
        self.obj_id = obj_id
        self.vdom = vdom

        # Set and used only for objects that are configured via global context vs vdom context
        # these allow to determine if cli should use "config system global" or just configure the object
        # depending on whether vdom is enabled
        self.vdom_enabled = None
        self.is_global = None

        # Map of instance attribute names to fg attribute names
        self._data_attrs = {}
        # In CLI config output some attributes in data_attrs may not be needed. So set which to ignore for CLI
        self._cli_ignore_attrs = {}

        # For use in string output of object such as dunder methods __str__ and __repr__
        self._obj_to_str = f'obj_id={self.obj_id}, vdom={self.vdom}'

    # Instance to string dunder methods
    def __str__(self):
        return self._obj_to_str

    def __repr__(self):
        return self._obj_to_str

    # Property Methods
    @property
    def vdom(self):
        return self._vdom

    @vdom.setter
    def vdom(self, vdom):
        """ Validate the vdom argument for FortiGate consumption, if acceptable set to self.vdom

        Verify that vdom argument is non-empty string, with no whitespace between 1 and 31 chars.  If valid then set
        self.vdom = vdom

        Args:
            vdom (str): FortiGate VDOM name that the instance object should be tied to

        Returns:
            None
        """
        if vdom is None:
            self._vdom = None

        else:
            if isinstance(vdom, str):
                # vdom names cannot have spaces so check for spaces and throw error if there are
                for char in vdom:
                    if str.isspace(char):
                        raise ValueError("'vdom', when set, must be str with no whitespace")

                # Check vdom name string length meets FG requirements
                if 1 <= len(vdom) <= 31:
                    self._vdom = vdom
                else:
                    raise ValueError("'vdom', when set, must be a str between 1 and 31 chars")
            else:
                raise ValueError("'vdom', when set, must be a str between 1 and 31")

    # API Config Methods
    def get_api_config_add(self):
        """ Get FortiGate API configuration for adding(post) self to FortiGate via API using Fortinet's ftntlib library

        Based on currently set instance attributes this method will build and return a FortiGate API configuration
        parameters to be used in making Rest API post call to FortiGate using ftntlib.  (ftntlib is a library for
        simplifying REST API calls to FortiOS and can be downloaded from the Fortinet Developer Network at
        https://fndn.fortinet.net)

        Args:
            self: the current instance object

        Returns:
            A dictionary mapping keys to corresponding ftnlib API attributes  for ftntlib REST API methods.

            example:
                {'api': 'cmdb',
                'path': 'vpn.ipsec',
                'name': 'phase1-interface',
                'mkey': 'vpn1',
                'action': None,
                'data': {},
                'parameters': {'vdom': 'vdom1'}}
        """

        conf = {'api': self.API, 'path': self.API_PATH, 'name': self.API_NAME, 'mkey': self.API_MKEY, 'action': None}
        data = {}
        params = {}

        # Set the VDOM, if necessary
        if self.is_global:
            pass
        elif self.vdom:
            if self.vdom == 'global':
                pass
            else:
                params.update({'vdom': self.vdom})

        for inst_attr, fg_attr in self._data_attrs.items():
            if getattr(self, inst_attr): data.update({fg_attr: getattr(self, inst_attr)})

        # Add data and parameter dictionaries to conf dictionary
        conf.update({'data': data})
        conf.update({'parameters': params})

        return conf

    def get_api_config_update(self):
        """ Get FortiGate API configuration for updating(put) self to FortiGate via API using Fortinet's ftntlib library

        Based on current instance's attributes this method will build and return a FortiGate API configuration
        parameters to be used in making Rest API put call to FortiGate using ftntlib.  (ftntlib is a library for
        simplifying REST API calls to FortiOS and can be downloaded from the Fortinet Developer Network at
        https://fndn.fortinet.net)

        Args:
            self: the current instance object

        Returns:
            A dictionary mapping keys to corresponding ftnlib API attributes  for ftntlib REST API methods.

            example:
                {'api': 'cmdb',
                'path': 'vpn.ipsec',
                'name': 'phase1-interface',
                'mkey': 'vpn1',
                'action': None,
                'data': {},
                'parameters': {'vdom': 'vdom1'}}
        """

        # Need to set mkey to interface name when doing updates (puts) or deletes
        self.API_MKEY = self.obj_id

        conf = self.get_api_config_add()
        return conf

    def get_api_config_del(self):
        """ Get FortiGate API configuration for deleting(delete) self from FortiGate via API using Fortinet ftntlib lib

        Based on current instance's attributes this method will build and return a FortiGate API configuration
        parameters to be used in making Rest API delete call to FortiGate using ftntlib.  (ftntlib is a library for
        simplifying REST API calls to FortiOS and can be downloaded from the Fortinet Developer Network at
        https://fndn.fortinet.net)

        Args:
            self: the current instance object

        Returns:
            A dictionary mapping keys to corresponding ftnlib API attributes  for ftntlib REST API methods.

            example:
                {'api': 'cmdb',
                'path': 'vpn.ipsec',
                'name': 'phase1-interface',
                'mkey': 'vpn1',
                'action': None,
                'data': {},
                'parameters': {'vdom': 'vdom1'}}

        """
        conf = {'api': self.API, 'path': self.API_PATH, 'name': self.API_NAME, 'mkey': self.API_MKEY, 'action': None}
        data = {}
        params = {}

        # Set the VDOM, if necessary
        if self.vdom:
            if self.vdom == 'global':
                pass
            else:
                params.update({'vdom': self.vdom})

        if self.obj_id:
            # Set the mkey value to interface name and updated other vars
            conf['mkey'] = self.obj_id
            conf.update({'data': data})
            conf.update({'parameters': params})

        else:
            raise Exception("\"name\" must be set in order get or delete an existing policy")

        return conf

    def get_api_config_get(self):
        """ Get FortiGate API configuration for getting(get) self from FortiGate via API using Fortinet ftntlib lib

            Based on current instance's attributes this method will build and return a FortiGate API configuration
            parameters to be used in making Rest API get call to FortiGate using ftntlib.  (ftntlib is a library for
            simplifying REST API calls to FortiOS and can be downloaded from the Fortinet Developer Network at
            https://fndn.fortinet.net)

        Args:
            self: the current instance object

        Returns:
            A dictionary mapping keys to corresponding ftnlib API attributes  for ftntlib REST API methods.

            example:
                {'api': 'cmdb',
                'path': 'vpn.ipsec',
                'name': 'phase1-interface',
                'mkey': 'vpn1',
                'action': None,
                'data': {},
                'parameters': {'vdom': 'vdom1'}}
            """
        conf = self.get_api_config_del()
        return conf

    # CLI Config Methods
    def get_cli_config_add(self):
        """ Get FortiGate CLI configuration for adding self to FortiGate via CLI

            Based on currently set instance attributes this method will build and return a FortiGate CLI configuration
            snippet for the current instance object.

            Args:
                self: the current instance object

            Returns:
                A FortiGate CLI configuration snippet representing self to be use for configuring new FG object.
            """

        conf = ''

        # start vdom or global config
        if self.vdom:
            if self.vdom == 'global' and self.vdom_enabled:
                conf += "config global\n"
            else:
                conf += "config vdom\n"
                conf += f" edit {self.vdom} \n"

        # Config object's cli path
        conf += f"{self.CLI_PATH}\n"

        # Edit obj_id
        conf += f"  edit \"{self.obj_id}\" \n"

        # For every attr defined in the data_attrs dictionary, if the dictionary value is true then add it to the
        # configuration.  Otherwise skip it.
        for inst_attr, fg_attr in self._data_attrs.items():

            # Check for cli attribute in ignore list and skip if contained
            if inst_attr in self._cli_ignore_attrs: continue

            # get the value of an attribute based on the text name of the attribute in data_attrs dictionary
            config_attr = getattr(self, inst_attr)

            # need to convert lists which are used for api, to strings for cli output
            if isinstance(config_attr, list):
                str_items = ''

                # if the config item is a list, then get the dictionaries from that list, pull the value and assign
                # the value to a string that will be used as the config parameters in the cli config output
                for item in config_attr:
                    if isinstance(item, dict):
                        if item['name']:
                            str_items += f"{item.get('name')} "
                        else:
                            raise Exception(f"unrecognized key name for dictionary list: {item.keys()}")

                conf += f"    set {fg_attr} \"{str_items}\"\n"
            else:
                if getattr(self, inst_attr): conf += "    set {fg_attr} \"{config_attr}\"\n"

        # End obj_id config
        conf += "  end\nend\n"

        # End vdom or global config
        if self.vdom:
            if self.vdom == 'global' and self.vdom_enabled:
                conf += "end\n"
            elif self.vdom == 'global':
                pass
            else:
                conf += "end\n"

        return conf

    def get_cli_config_update(self):
        """ Get FortiGate CLI configuration for updating self to FortiGate via CLI

        Based on currently set instance attributes this method will build and return a FortiGate CLI configuration
        snippet for the current instance object.

        Args:
            self: the current instance object

        Returns:
            A FortiGate CLI configuration snippet representing self to be use for updating existing FG object
        """
        conf = self.get_cli_config_add()
        return conf

    def get_cli_config_del(self):
        """ Get FortiGate CLI configuration for deleting self from FortiGate via CLI

            Based on currently set instance attributes this method will build and return a FortiGate CLI configuration
            snippet for the current instance object.

            Args:
                self: the current instance object

            Returns:
                A FortiGate CLI configuration snippet representing self to be use for configuring new FG object
            """
        conf = ''
        if self.obj_id:

            # start vdom or global config
            if self.vdom:
                if self.vdom == 'global' and self.vdom_enabled:
                    conf += "config global\n"
                else:
                    conf += "config vdom\n"
                    conf += f" edit {self.vdom} \n"

            conf += f"{self.CLI_PATH}\n"
            conf += f"  delete {self.obj_id}\n"
            conf += "end\n"

            # End vdom or global config
            if self.vdom:
                if self.vdom == 'global' and self.vdom_enabled:
                    conf += "end\n"
                elif self.vdom == 'global':
                    pass
                else:
                    conf += "end\n"

            return conf
        else:
            raise Exception("'obj_id' must be set in order to configure it for delete")
